@@ -9,13 +9,8 @@
 
 **********************************************************************/
 
-int rb_encdb_replicate(const char *alias, const char *orig);
-int rb_encdb_alias(const char *alias, const char *orig);
-int rb_encdb_dummy(const char *name);
-void rb_encdb_declare(const char *name);
-void rb_enc_set_base(const char *name, const char *orig);
-int rb_enc_set_dummy(int index);
-void rb_encdb_set_unicode(int index);
+#include "internal.h"
+
 #define ENC_REPLICATE(name, orig) rb_encdb_replicate((name), (orig))
 #define ENC_ALIAS(name, orig) rb_encdb_alias((name), (orig))
 #define ENC_DUMMY(name) rb_encdb_dummy(name)
