@@ -17,11 +17,12 @@ else
 end
 if ok
   have_header("sys/ioctl.h") if hdr
-  have_func("rb_funcallv")
-  have_func("rb_sym2str")
   # rb_check_hash_type: 1.9.3
   # rb_io_get_write_io: 1.9.1
   # rb_cloexec_open: 2.0.0
+  # rb_funcallv: 2.1.0
+  # RARRAY_CONST_PTR: 2.1.0
+  # rb_sym2str: 2.2.0
   $defs << "-D""ENABLE_IO_GETPASS=1"
   create_makefile("io/console") {|conf|
     conf << "\n""VK_HEADER = #{vk_header}\n"
