@@ -3,6 +3,7 @@
 #include <time.h>
 
 int rsock_cmsg_cloexec_state = -1; /* <0: unknown, 0: ignored, >0: working */
+static VALUE sym_wait_readable, sym_wait_writable;
 
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
 static VALUE rb_cAncillaryData;
@@ -206,7 +207,7 @@ ancillary_s_unix_rights(int argc, VALUE *argv, VALUE klass)
     str = rb_str_buf_new(sizeof(int) * argc);
 
     for (i = 0 ; i < argc; i++) {
-        VALUE obj = RARRAY_PTR(ary)[i];
+        VALUE obj = RARRAY_AREF(ary, i);
         rb_io_t *fptr;
         int fd;
         GetOpenFile(obj, fptr);
@@ -359,6 +360,8 @@ ancillary_timestamp(VALUE self)
  *
  * The size and endian is dependent on the host.
  *
+ *   require 'socket'
+ *
  *   p Socket::AncillaryData.int(:UNIX, :SOCKET, :RIGHTS, STDERR.fileno)
  *   #=> #<Socket::AncillaryData: UNIX SOCKET RIGHTS 2>
  */
@@ -989,7 +992,7 @@ ancillary_inspect(VALUE self)
 
         vtype = ip_cmsg_type_to_sym(level, type);
         if (SYMBOL_P(vtype))
-            rb_str_catf(ret, " %s", rb_id2name(SYM2ID(vtype)));
+            rb_str_catf(ret, " %"PRIsVALUE, rb_sym2str(vtype));
         else
             rb_str_catf(ret, " cmsg_type:%d", type);
     }
@@ -1127,16 +1130,17 @@ rb_sendmsg(int fd, const struct msghdr *msg, int flags)
 }
 
 static VALUE
-bsock_sendmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
+bsock_sendmsg_internal(VALUE sock, VALUE data, VALUE vflags,
+		       VALUE dest_sockaddr, VALUE controls, VALUE ex,
+		       int nonblock)
 {
     rb_io_t *fptr;
-    VALUE data, vflags, dest_sockaddr;
-    int controls_num;
     struct msghdr mh;
     struct iovec iov;
+    VALUE tmp;
+    int controls_num;
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
     VALUE controls_str = 0;
-    VALUE *controls_ptr = NULL;
     int family;
 #endif
     int flags;
@@ -1144,27 +1148,22 @@ bsock_sendmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
 
     GetOpenFile(sock, fptr);
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-    family = rsock_getfamily(fptr->fd);
-#endif
-
-    data = vflags = dest_sockaddr = Qnil;
-
-    if (argc == 0)
-        rb_raise(rb_eArgError, "mesg argument required");
-    data = argv[0];
-    if (1 < argc) vflags = argv[1];
-    if (2 < argc) dest_sockaddr = argv[2];
-    controls_num = 3 < argc ? argc - 3 : 0;
-#if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-    if (3 < argc) { controls_ptr = &argv[3]; }
+    family = rsock_getfamily(fptr);
 #endif
 
     StringValue(data);
+    tmp = rb_str_tmp_frozen_acquire(data);
+
+    if (!RB_TYPE_P(controls, T_ARRAY)) {
+	controls = rb_ary_new();
+    }
+    controls_num = RARRAY_LENINT(controls);
 
     if (controls_num) {
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
 	int i;
 	size_t last_pad = 0;
+	const VALUE *controls_ptr = RARRAY_CONST_PTR(controls);
 #if defined(__NetBSD__)
         int last_level = 0;
         int last_type = 0;
@@ -1239,6 +1238,7 @@ bsock_sendmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
                 rb_str_set_len(controls_str, RSTRING_LEN(controls_str)-last_pad);
 #endif
 	}
+	RB_GC_GUARD(controls);
 #else
 	rb_raise(rb_eNotImpError, "control message for sendmsg is unimplemented");
 #endif
@@ -1263,98 +1263,63 @@ bsock_sendmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
     }
     mh.msg_iovlen = 1;
     mh.msg_iov = &iov;
-    iov.iov_base = RSTRING_PTR(data);
-    iov.iov_len = RSTRING_LEN(data);
+    iov.iov_base = RSTRING_PTR(tmp);
+    iov.iov_len = RSTRING_LEN(tmp);
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
     if (controls_str) {
         mh.msg_control = RSTRING_PTR(controls_str);
         mh.msg_controllen = RSTRING_SOCKLEN(controls_str);
     }
-    else {
-        mh.msg_control = NULL;
-        mh.msg_controllen = 0;
-    }
 #endif
 
     rb_io_check_closed(fptr);
-    if (nonblock)
+    if (nonblock && !MSG_DONTWAIT_RELIABLE)
         rb_io_set_nonblock(fptr);
 
     ss = rb_sendmsg(fptr->fd, &mh, flags);
 
     if (ss == -1) {
+	int e;
         if (!nonblock && rb_io_wait_writable(fptr->fd)) {
             rb_io_check_closed(fptr);
             goto retry;
         }
-        if (nonblock && (errno == EWOULDBLOCK || errno == EAGAIN))
-            rb_readwrite_sys_fail(RB_IO_WAIT_WRITABLE, "sendmsg(2) would block");
-	rb_sys_fail("sendmsg(2)");
+	e = errno;
+	if (nonblock && (e == EWOULDBLOCK || e == EAGAIN)) {
+	    if (ex == Qfalse) {
+		return sym_wait_writable;
+	    }
+	    rb_readwrite_syserr_fail(RB_IO_WAIT_WRITABLE, e,
+				     "sendmsg(2) would block");
+	}
+	rb_syserr_fail(e, "sendmsg(2)");
     }
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
     RB_GC_GUARD(controls_str);
 #endif
+    rb_str_tmp_frozen_release(data, tmp);
 
     return SSIZET2NUM(ss);
 }
 #endif
 
 #if defined(HAVE_SENDMSG)
-/*
- * call-seq:
- *    basicsocket.sendmsg(mesg, flags=0, dest_sockaddr=nil, *controls) => numbytes_sent
- *
- * sendmsg sends a message using sendmsg(2) system call in blocking manner.
- *
- * _mesg_ is a string to send.
- *
- * _flags_ is bitwise OR of MSG_* constants such as Socket::MSG_OOB.
- *
- * _dest_sockaddr_ is a destination socket address for connection-less socket.
- * It should be a sockaddr such as a result of Socket.sockaddr_in.
- * An Addrinfo object can be used too.
- *
- * _controls_ is a list of ancillary data.
- * The element of _controls_ should be Socket::AncillaryData or
- * 3-elements array.
- * The 3-element array should contains cmsg_level, cmsg_type and data.
- *
- * The return value, _numbytes_sent_ is an integer which is the number of bytes sent.
- *
- * sendmsg can be used to implement send_io as follows:
- *
- *   # use Socket::AncillaryData.
- *   ancdata = Socket::AncillaryData.int(:UNIX, :SOCKET, :RIGHTS, io.fileno)
- *   sock.sendmsg("a", 0, nil, ancdata)
- *
- *   # use 3-element array.
- *   ancdata = [:SOCKET, :RIGHTS, [io.fileno].pack("i!")]
- *   sock.sendmsg("\0", 0, nil, ancdata)
- *
- */
 VALUE
-rsock_bsock_sendmsg(int argc, VALUE *argv, VALUE sock)
+rsock_bsock_sendmsg(VALUE sock, VALUE data, VALUE flags, VALUE dest_sockaddr,
+		    VALUE controls)
 {
-    return bsock_sendmsg_internal(argc, argv, sock, 0);
+    return bsock_sendmsg_internal(sock, data, flags, dest_sockaddr, controls,
+				  Qtrue, 0);
 }
 #endif
 
 #if defined(HAVE_SENDMSG)
-/*
- * call-seq:
- *    basicsocket.sendmsg_nonblock(mesg, flags=0, dest_sockaddr=nil, *controls) => numbytes_sent
- *
- * sendmsg_nonblock sends a message using sendmsg(2) system call in non-blocking manner.
- *
- * It is similar to BasicSocket#sendmsg
- * but the non-blocking flag is set before the system call
- * and it doesn't retry the system call.
- *
- */
 VALUE
-rsock_bsock_sendmsg_nonblock(int argc, VALUE *argv, VALUE sock)
+rsock_bsock_sendmsg_nonblock(VALUE sock, VALUE data, VALUE flags,
+			     VALUE dest_sockaddr, VALUE controls, VALUE ex)
 {
-    return bsock_sendmsg_internal(argc, argv, sock, 1);
+    return bsock_sendmsg_internal(sock, data, flags, dest_sockaddr,
+				  controls, ex, 1);
 }
 #endif
 
@@ -1482,18 +1447,18 @@ make_io_for_unix_rights(VALUE ctl, struct cmsghdr *cmh, char *msg_end)
 #endif
 
 static VALUE
-bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
+bsock_recvmsg_internal(VALUE sock,
+		VALUE vmaxdatlen, VALUE vflags, VALUE vmaxctllen,
+		VALUE scm_rights, VALUE ex, int nonblock)
 {
     rb_io_t *fptr;
-    VALUE vmaxdatlen, vmaxctllen, vflags;
-    VALUE vopts;
     int grow_buffer;
     size_t maxdatlen;
     int flags, orig_flags;
     struct msghdr mh;
     struct iovec iov;
     union_sockaddr namebuf;
-    char datbuf0[4096], *datbuf;
+    char *datbuf;
     VALUE dat_str = Qnil;
     VALUE ret;
     ssize_t ss;
@@ -1501,27 +1466,20 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
     struct cmsghdr *cmh;
     size_t maxctllen;
-    union {
-        char bytes[4096];
-        struct cmsghdr align;
-    } ctlbuf0;
     char *ctlbuf;
     VALUE ctl_str = Qnil;
     int family;
     int gc_done = 0;
 #endif
 
-
-    rb_scan_args(argc, argv, "03:", &vmaxdatlen, &vflags, &vmaxctllen, &vopts);
-
-    maxdatlen = NIL_P(vmaxdatlen) ? sizeof(datbuf0) : NUM2SIZET(vmaxdatlen);
+    maxdatlen = NIL_P(vmaxdatlen) ? 4096 : NUM2SIZET(vmaxdatlen);
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-    maxctllen = NIL_P(vmaxctllen) ? sizeof(ctlbuf0) : NUM2SIZET(vmaxctllen);
+    maxctllen = NIL_P(vmaxctllen) ? 4096 : NUM2SIZET(vmaxctllen);
 #else
     if (!NIL_P(vmaxctllen))
         rb_raise(rb_eArgError, "control message not supported");
 #endif
-    flags = NIL_P(vflags) ? 0 : NUM2INT(vflags);
+    flags = NUM2INT(vflags);
 #ifdef MSG_DONTWAIT
     if (nonblock)
         flags |= MSG_DONTWAIT;
@@ -1531,7 +1489,7 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
     grow_buffer = NIL_P(vmaxdatlen) || NIL_P(vmaxctllen);
 
     request_scm_rights = 0;
-    if (!NIL_P(vopts) && RTEST(rb_hash_aref(vopts, ID2SYM(rb_intern("scm_rights")))))
+    if (RTEST(scm_rights))
         request_scm_rights = 1;
 #if !defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
     if (request_scm_rights)
@@ -1556,26 +1514,18 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
 #endif
 
   retry:
-    if (maxdatlen <= sizeof(datbuf0))
-        datbuf = datbuf0;
-    else {
-        if (NIL_P(dat_str))
-            dat_str = rb_str_tmp_new(maxdatlen);
-        else
-            rb_str_resize(dat_str, maxdatlen);
-        datbuf = RSTRING_PTR(dat_str);
-    }
+    if (NIL_P(dat_str))
+	dat_str = rb_str_tmp_new(maxdatlen);
+    else
+	rb_str_resize(dat_str, maxdatlen);
+    datbuf = RSTRING_PTR(dat_str);
 
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-    if (maxctllen <= sizeof(ctlbuf0))
-        ctlbuf = ctlbuf0.bytes;
-    else {
-        if (NIL_P(ctl_str))
-            ctl_str = rb_str_tmp_new(maxctllen);
-        else
-            rb_str_resize(ctl_str, maxctllen);
-        ctlbuf = RSTRING_PTR(ctl_str);
-    }
+    if (NIL_P(ctl_str))
+	ctl_str = rb_str_tmp_new(maxctllen);
+    else
+	rb_str_resize(ctl_str, maxctllen);
+    ctlbuf = RSTRING_PTR(ctl_str);
 #endif
 
     memset(&mh, 0, sizeof(mh));
@@ -1598,20 +1548,26 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
         flags |= MSG_PEEK;
 
     rb_io_check_closed(fptr);
-    if (nonblock)
+    if (nonblock && !MSG_DONTWAIT_RELIABLE)
         rb_io_set_nonblock(fptr);
 
     ss = rb_recvmsg(fptr->fd, &mh, flags);
 
     if (ss == -1) {
+	int e;
         if (!nonblock && rb_io_wait_readable(fptr->fd)) {
             rb_io_check_closed(fptr);
             goto retry;
         }
-        if (nonblock && (errno == EWOULDBLOCK || errno == EAGAIN))
-            rb_readwrite_sys_fail(RB_IO_WAIT_READABLE, "recvmsg(2) would block");
+	e = errno;
+	if (nonblock && (e == EWOULDBLOCK || e == EAGAIN)) {
+            if (ex == Qfalse) {
+                return sym_wait_readable;
+            }
+	    rb_readwrite_syserr_fail(RB_IO_WAIT_READABLE, e, "recvmsg(2) would block");
+        }
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-        if (!gc_done && (errno == EMFILE || errno == EMSGSIZE)) {
+	if (!gc_done && (e == EMFILE || e == EMSGSIZE)) {
           /*
            * When SCM_RIGHTS hit the file descriptors limit:
            * - Linux 2.6.18 causes success with MSG_CTRUNC
@@ -1623,19 +1579,23 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
             gc_done = 1;
 	    goto retry;
         }
+#else
+	if (NIL_P(vmaxdatlen) && grow_buffer && e == EMSGSIZE)
+	    ss = (ssize_t)iov.iov_len;
+	else
 #endif
-	rb_sys_fail("recvmsg(2)");
+	rb_syserr_fail(e, "recvmsg(2)");
     }
 
     if (grow_buffer) {
 	int grown = 0;
-#if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-        if (NIL_P(vmaxdatlen) && (mh.msg_flags & MSG_TRUNC)) {
+	if (NIL_P(vmaxdatlen) && ss != -1 && ss == (ssize_t)iov.iov_len) {
             if (SIZE_MAX/2 < maxdatlen)
                 rb_raise(rb_eArgError, "max data length too big");
 	    maxdatlen *= 2;
 	    grown = 1;
 	}
+#if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
         if (NIL_P(vmaxctllen) && (mh.msg_flags & MSG_CTRUNC)) {
 #define BIG_ENOUGH_SPACE 65536
             if (BIG_ENOUGH_SPACE < maxctllen &&
@@ -1655,13 +1615,6 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
             }
 #undef BIG_ENOUGH_SPACE
 	}
-#else
-	if (NIL_P(vmaxdatlen) && ss != -1 && ss == (ssize_t)iov.iov_len) {
-            if (SIZE_MAX/2 < maxdatlen)
-                rb_raise(rb_eArgError, "max data length too big");
-	    maxdatlen *= 2;
-	    grown = 1;
-	}
 #endif
 	if (grown) {
             rsock_discard_cmsg_resource(&mh, (flags & MSG_PEEK) != 0);
@@ -1678,10 +1631,9 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
     }
 
     if (NIL_P(dat_str))
-        dat_str = rb_tainted_str_new(datbuf, ss);
+        dat_str = rb_str_new(datbuf, ss);
     else {
         rb_str_resize(dat_str, ss);
-        OBJ_TAINT(dat_str);
 	rb_obj_reveal(dat_str, rb_cString);
     }
 
@@ -1695,7 +1647,7 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
 			 );
 
 #if defined(HAVE_STRUCT_MSGHDR_MSG_CONTROL)
-    family = rsock_getfamily(fptr->fd);
+    family = rsock_getfamily(fptr);
     if (mh.msg_controllen) {
 	char *msg_end = (char *)mh.msg_control + mh.msg_controllen;
         for (cmh = CMSG_FIRSTHDR(&mh); cmh != NULL; cmh = CMSG_NXTHDR(&mh, cmh)) {
@@ -1707,7 +1659,7 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
             }
             ctl_end = (char*)cmh + cmh->cmsg_len;
 	    clen = (ctl_end <= msg_end ? ctl_end : msg_end) - (char*)CMSG_DATA(cmh);
-            ctl = ancdata_new(family, cmh->cmsg_level, cmh->cmsg_type, rb_tainted_str_new((char*)CMSG_DATA(cmh), clen));
+            ctl = ancdata_new(family, cmh->cmsg_level, cmh->cmsg_type, rb_str_new((char*)CMSG_DATA(cmh), clen));
             if (request_scm_rights)
                 make_io_for_unix_rights(ctl, cmh, msg_end);
             else
@@ -1723,82 +1675,21 @@ bsock_recvmsg_internal(int argc, VALUE *argv, VALUE sock, int nonblock)
 #endif
 
 #if defined(HAVE_RECVMSG)
-/*
- * call-seq:
- *    basicsocket.recvmsg(maxmesglen=nil, flags=0, maxcontrollen=nil, opts={}) => [mesg, sender_addrinfo, rflags, *controls]
- *
- * recvmsg receives a message using recvmsg(2) system call in blocking manner.
- *
- * _maxmesglen_ is the maximum length of mesg to receive.
- *
- * _flags_ is bitwise OR of MSG_* constants such as Socket::MSG_PEEK.
- *
- * _maxcontrollen_ is the maximum length of controls (ancillary data) to receive.
- *
- * _opts_ is option hash.
- * Currently :scm_rights=>bool is the only option.
- *
- * :scm_rights option specifies that application expects SCM_RIGHTS control message.
- * If the value is nil or false, application don't expects SCM_RIGHTS control message.
- * In this case, recvmsg closes the passed file descriptors immediately.
- * This is the default behavior.
- *
- * If :scm_rights value is neither nil nor false, application expects SCM_RIGHTS control message.
- * In this case, recvmsg creates IO objects for each file descriptors for
- * Socket::AncillaryData#unix_rights method.
- *
- * The return value is 4-elements array.
- *
- * _mesg_ is a string of the received message.
- *
- * _sender_addrinfo_ is a sender socket address for connection-less socket.
- * It is an Addrinfo object.
- * For connection-oriented socket such as TCP, sender_addrinfo is platform dependent.
- *
- * _rflags_ is a flags on the received message which is bitwise OR of MSG_* constants such as Socket::MSG_TRUNC.
- * It will be nil if the system uses 4.3BSD style old recvmsg system call.
- *
- * _controls_ is ancillary data which is an array of Socket::AncillaryData objects such as:
- *
- *   #<Socket::AncillaryData: AF_UNIX SOCKET RIGHTS 7>
- *
- * _maxmesglen_ and _maxcontrollen_ can be nil.
- * In that case, the buffer will be grown until the message is not truncated.
- * Internally, MSG_PEEK is used and MSG_TRUNC/MSG_CTRUNC are checked.
- *
- * recvmsg can be used to implement recv_io as follows:
- *
- *   mesg, sender_sockaddr, rflags, *controls = sock.recvmsg(:scm_rights=>true)
- *   controls.each {|ancdata|
- *     if ancdata.cmsg_is?(:SOCKET, :RIGHTS)
- *       return ancdata.unix_rights[0]
- *     end
- *   }
- *
- */
 VALUE
-rsock_bsock_recvmsg(int argc, VALUE *argv, VALUE sock)
+rsock_bsock_recvmsg(VALUE sock, VALUE dlen, VALUE flags, VALUE clen,
+		    VALUE scm_rights)
 {
-    return bsock_recvmsg_internal(argc, argv, sock, 0);
+    VALUE ex = Qtrue;
+    return bsock_recvmsg_internal(sock, dlen, flags, clen, scm_rights, ex, 0);
 }
 #endif
 
 #if defined(HAVE_RECVMSG)
-/*
- * call-seq:
- *    basicsocket.recvmsg_nonblock(maxdatalen=nil, flags=0, maxcontrollen=nil, opts={}) => [data, sender_addrinfo, rflags, *controls]
- *
- * recvmsg receives a message using recvmsg(2) system call in non-blocking manner.
- *
- * It is similar to BasicSocket#recvmsg
- * but non-blocking flag is set before the system call
- * and it doesn't retry the system call.
- *
- */
 VALUE
-rsock_bsock_recvmsg_nonblock(int argc, VALUE *argv, VALUE sock)
+rsock_bsock_recvmsg_nonblock(VALUE sock, VALUE dlen, VALUE flags, VALUE clen,
+			     VALUE scm_rights, VALUE ex)
 {
-    return bsock_recvmsg_internal(argc, argv, sock, 1);
+    return bsock_recvmsg_internal(sock, dlen, flags, clen, scm_rights, ex, 1);
 }
 #endif
 
@@ -1839,4 +1730,7 @@ rsock_init_ancdata(void)
     rb_define_method(rb_cAncillaryData, "ipv6_pktinfo_addr", ancillary_ipv6_pktinfo_addr, 0);
     rb_define_method(rb_cAncillaryData, "ipv6_pktinfo_ifindex", ancillary_ipv6_pktinfo_ifindex, 0);
 #endif
+#undef rb_intern
+    sym_wait_readable = ID2SYM(rb_intern("wait_readable"));
+    sym_wait_writable = ID2SYM(rb_intern("wait_writable"));
 }
