@@ -83,7 +83,7 @@ ole_method_sub(VALUE self, ITypeInfo *pOwnerTypeInfo, ITypeInfo *pTypeInfo, VALU
     VALUE method = Qnil;
     hr = OLE_GET_TYPEATTR(pTypeInfo, &pTypeAttr);
     if (FAILED(hr)) {
-        ole_raise(hr, eWIN32OLERuntimeError, "failed to GetTypeAttr");
+        ole_raise(hr, eWIN32OLEQueryInterfaceError, "failed to GetTypeAttr");
     }
     for(i = 0; i < pTypeAttr->cFuncs && method == Qnil; i++) {
         hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, i, &pFuncDesc);
@@ -119,7 +119,7 @@ ole_methods_from_typeinfo(ITypeInfo *pTypeInfo, int mask)
     VALUE methods = rb_ary_new();
     hr = OLE_GET_TYPEATTR(pTypeInfo, &pTypeAttr);
     if (FAILED(hr)) {
-        ole_raise(hr, eWIN32OLERuntimeError, "failed to GetTypeAttr");
+        ole_raise(hr, eWIN32OLEQueryInterfaceError, "failed to GetTypeAttr");
     }
 
     ole_methods_sub(0, pTypeInfo, methods, mask);
@@ -148,7 +148,7 @@ olemethod_from_typeinfo(VALUE self, ITypeInfo *pTypeInfo, VALUE name)
     VALUE method = Qnil;
     hr = OLE_GET_TYPEATTR(pTypeInfo, &pTypeAttr);
     if (FAILED(hr)) {
-        ole_raise(hr, eWIN32OLERuntimeError, "failed to GetTypeAttr");
+        ole_raise(hr, eWIN32OLEQueryInterfaceError, "failed to GetTypeAttr");
     }
     method = ole_method_sub(self, 0, pTypeInfo, name);
     if (method != Qnil) {
@@ -179,7 +179,7 @@ ole_methods_sub(ITypeInfo *pOwnerTypeInfo, ITypeInfo *pTypeInfo, VALUE methods,
     WORD i;
     hr = OLE_GET_TYPEATTR(pTypeInfo, &pTypeAttr);
     if (FAILED(hr)) {
-        ole_raise(hr, eWIN32OLERuntimeError, "failed to GetTypeAttr");
+        ole_raise(hr, eWIN32OLEQueryInterfaceError, "failed to GetTypeAttr");
     }
     for(i = 0; i < pTypeAttr->cFuncs; i++) {
         hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, i, &pFuncDesc);
@@ -283,7 +283,7 @@ folemethod_initialize(VALUE self, VALUE oletype, VALUE method)
 }
 
 /*
- *  call-seq
+ *  call-seq:
  *     WIN32OLE_METHOD#name
  *
  *  Returns the name of the method.
@@ -308,7 +308,7 @@ ole_method_return_type(ITypeInfo *pTypeInfo, UINT method_index)
 
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if (FAILED(hr))
-        ole_raise(hr, eWIN32OLERuntimeError, "failed to GetFuncDesc");
+        ole_raise(hr, eWIN32OLEQueryInterfaceError, "failed to GetFuncDesc");
 
     type = ole_typedesc2val(pTypeInfo, &(pFuncDesc->elemdescFunc.tdesc), Qnil);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
@@ -344,7 +344,7 @@ ole_method_return_vtype(ITypeInfo *pTypeInfo, UINT method_index)
     if (FAILED(hr))
         ole_raise(hr, eWIN32OLERuntimeError, "failed to GetFuncDesc");
 
-    vvt = INT2FIX(pFuncDesc->elemdescFunc.tdesc.vt);
+    vvt = RB_INT2FIX(pFuncDesc->elemdescFunc.tdesc.vt);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return vvt;
 }
@@ -410,7 +410,7 @@ ole_method_invkind(ITypeInfo *pTypeInfo, UINT method_index)
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if(FAILED(hr))
         ole_raise(hr, eWIN32OLERuntimeError, "failed to GetFuncDesc");
-    invkind = INT2FIX(pFuncDesc->invkind);
+    invkind = RB_INT2FIX(pFuncDesc->invkind);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return invkind;
 }
@@ -420,16 +420,16 @@ ole_method_invoke_kind(ITypeInfo *pTypeInfo, UINT method_index)
 {
     VALUE type = rb_str_new2("UNKNOWN");
     VALUE invkind = ole_method_invkind(pTypeInfo, method_index);
-    if((FIX2INT(invkind) & INVOKE_PROPERTYGET) &&
-       (FIX2INT(invkind) & INVOKE_PROPERTYPUT) ) {
+    if((RB_FIX2INT(invkind) & INVOKE_PROPERTYGET) &&
+       (RB_FIX2INT(invkind) & INVOKE_PROPERTYPUT) ) {
         type = rb_str_new2("PROPERTY");
-    } else if(FIX2INT(invkind) & INVOKE_PROPERTYGET) {
+    } else if(RB_FIX2INT(invkind) & INVOKE_PROPERTYGET) {
         type =  rb_str_new2("PROPERTYGET");
-    } else if(FIX2INT(invkind) & INVOKE_PROPERTYPUT) {
+    } else if(RB_FIX2INT(invkind) & INVOKE_PROPERTYPUT) {
         type = rb_str_new2("PROPERTYPUT");
-    } else if(FIX2INT(invkind) & INVOKE_PROPERTYPUTREF) {
+    } else if(RB_FIX2INT(invkind) & INVOKE_PROPERTYPUTREF) {
         type = rb_str_new2("PROPERTYPUTREF");
-    } else if(FIX2INT(invkind) & INVOKE_FUNC) {
+    } else if(RB_FIX2INT(invkind) & INVOKE_FUNC) {
         type = rb_str_new2("FUNC");
     }
     return type;
@@ -712,7 +712,7 @@ ole_method_helpcontext(ITypeInfo *pTypeInfo, UINT method_index)
                                       &helpcontext, NULL);
     if (FAILED(hr))
         return Qnil;
-    return INT2FIX(helpcontext);
+    return RB_INT2FIX(helpcontext);
 }
 
 /*
@@ -741,7 +741,7 @@ ole_method_dispid(ITypeInfo *pTypeInfo, UINT method_index)
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if (FAILED(hr))
         return dispid;
-    dispid = INT2NUM(pFuncDesc->memid);
+    dispid = RB_INT2NUM(pFuncDesc->memid);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return dispid;
 }
@@ -772,7 +772,7 @@ ole_method_offset_vtbl(ITypeInfo *pTypeInfo, UINT method_index)
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if (FAILED(hr))
         return offset_vtbl;
-    offset_vtbl = INT2FIX(pFuncDesc->oVft);
+    offset_vtbl = RB_INT2FIX(pFuncDesc->oVft);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return offset_vtbl;
 }
@@ -803,7 +803,7 @@ ole_method_size_params(ITypeInfo *pTypeInfo, UINT method_index)
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if (FAILED(hr))
         return size_params;
-    size_params = INT2FIX(pFuncDesc->cParams);
+    size_params = RB_INT2FIX(pFuncDesc->cParams);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return size_params;
 }
@@ -835,7 +835,7 @@ ole_method_size_opt_params(ITypeInfo *pTypeInfo, UINT method_index)
     hr = pTypeInfo->lpVtbl->GetFuncDesc(pTypeInfo, method_index, &pFuncDesc);
     if (FAILED(hr))
         return size_opt_params;
-    size_opt_params = INT2FIX(pFuncDesc->cParamsOpt);
+    size_opt_params = RB_INT2FIX(pFuncDesc->cParamsOpt);
     pTypeInfo->lpVtbl->ReleaseFuncDesc(pTypeInfo, pFuncDesc);
     return size_opt_params;
 }
