@@ -2,16 +2,17 @@
 #
 #   irb/context.rb - irb context
 #   	$Release Version: 0.9.6$
-#   	$Revision: 58415 $
+#   	$Revision$
 #   	by Keiju ISHITSUKA(keiju@ruby-lang.org)
 #
 # --
 #
 #
 #
-require "irb/workspace"
-require "irb/inspector"
-require "irb/output-method"
+require_relative "workspace"
+require_relative "inspector"
+require_relative "input-method"
+require_relative "output-method"
 
 module IRB
   # A class that wraps the current state of the irb session, including the
@@ -21,10 +22,10 @@ module IRB
     #
     # The optional +input_method+ argument:
     #
-    # +nil+::     uses stdin or Readline
+    # +nil+::     uses stdin or Reidline or Readline
     # +String+::  uses a File
     # +other+::   uses this as InputMethod
-    def initialize(irb, workspace = nil, input_method = nil, output_method = nil)
+    def initialize(irb, workspace = nil, input_method = nil)
       @irb = irb
       if workspace
         @workspace = workspace
@@ -38,7 +39,21 @@ module IRB
       @rc = IRB.conf[:RC]
       @load_modules = IRB.conf[:LOAD_MODULES]
 
-      @use_readline = IRB.conf[:USE_READLINE]
+      if IRB.conf.has_key?(:USE_SINGLELINE)
+        @use_singleline = IRB.conf[:USE_SINGLELINE]
+      elsif IRB.conf.has_key?(:USE_READLINE) # backward compatibility
+        @use_singleline = IRB.conf[:USE_READLINE]
+      else
+        @use_singleline = nil
+      end
+      if IRB.conf.has_key?(:USE_MULTILINE)
+        @use_multiline = IRB.conf[:USE_MULTILINE]
+      elsif IRB.conf.has_key?(:USE_REIDLINE) # backward compatibility
+        @use_multiline = IRB.conf[:USE_REIDLINE]
+      else
+        @use_multiline = nil
+      end
+      @use_colorize = IRB.conf[:USE_COLORIZE]
       @verbose = IRB.conf[:VERBOSE]
       @io = nil
 
@@ -63,23 +78,42 @@ module IRB
 
       case input_method
       when nil
-        case use_readline?
+        @io = nil
+        case use_multiline?
         when nil
-          if (defined?(ReadlineInputMethod) && STDIN.tty? &&
-              IRB.conf[:PROMPT_MODE] != :INF_RUBY)
-            @io = ReadlineInputMethod.new
+          if STDIN.tty? && IRB.conf[:PROMPT_MODE] != :INF_RUBY && !use_singleline?
+            # Both of multiline mode and singleline mode aren't specified.
+            @io = ReidlineInputMethod.new
           else
-            @io = StdioInputMethod.new
+            @io = nil
           end
         when false
-          @io = StdioInputMethod.new
+          @io = nil
         when true
-          if defined?(ReadlineInputMethod)
-            @io = ReadlineInputMethod.new
+          @io = ReidlineInputMethod.new
+        end
+        unless @io
+          case use_singleline?
+          when nil
+            if (defined?(ReadlineInputMethod) && STDIN.tty? &&
+                IRB.conf[:PROMPT_MODE] != :INF_RUBY)
+              @io = ReadlineInputMethod.new
+            else
+              @io = nil
+            end
+          when false
+            @io = nil
+          when true
+            if defined?(ReadlineInputMethod)
+              @io = ReadlineInputMethod.new
+            else
+              @io = nil
+            end
           else
-            @io = StdioInputMethod.new
+            @io = nil
           end
         end
+        @io = StdioInputMethod.new unless @io
 
       when String
         @io = FileInputMethod.new(input_method)
@@ -90,17 +124,15 @@ module IRB
       end
       self.save_history = IRB.conf[:SAVE_HISTORY] if IRB.conf[:SAVE_HISTORY]
 
-      if output_method
-        @output_method = output_method
-      else
-        @output_method = StdioOutputMethod.new
-      end
-
       @echo = IRB.conf[:ECHO]
       if @echo.nil?
         @echo = true
       end
-      self.debug_level = IRB.conf[:DEBUG_LEVEL]
+
+      @echo_on_assignment = IRB.conf[:ECHO_ON_ASSIGNMENT]
+      if @echo_on_assignment.nil?
+        @echo_on_assignment = false
+      end
     end
 
     # The top-level workspace, see WorkSpace#main
@@ -116,9 +148,9 @@ module IRB
     attr_reader :thread
     # The current input method
     #
-    # Can be either StdioInputMethod, ReadlineInputMethod, FileInputMethod or
-    # other specified when the context is created. See ::new for more
-    # information on +input_method+.
+    # Can be either StdioInputMethod, ReadlineInputMethod,
+    # ReidlineInputMethod, FileInputMethod or other specified when the
+    # context is created. See ::new for more # information on +input_method+.
     attr_accessor :io
 
     # Current irb session
@@ -136,12 +168,18 @@ module IRB
     # +input_method+ passed to Context.new
     attr_accessor :irb_path
 
-    # Whether +Readline+ is enabled or not.
+    # Whether multiline editor mode is enabled or not.
     #
-    # A copy of the default <code>IRB.conf[:USE_READLINE]</code>
+    # A copy of the default <code>IRB.conf[:USE_MULTILINE]</code>
+    attr_reader :use_multiline
+    # Whether singleline editor mode is enabled or not.
     #
-    # See #use_readline= for more information.
-    attr_reader :use_readline
+    # A copy of the default <code>IRB.conf[:USE_SINGLELINE]</code>
+    attr_reader :use_singleline
+    # Whether colorization is enabled or not.
+    #
+    # A copy of the default <code>IRB.conf[:USE_COLORIZE]</code>
+    attr_reader :use_colorize
     # A copy of the default <code>IRB.conf[:INSPECT_MODE]</code>
     attr_reader :inspect_mode
 
@@ -164,17 +202,17 @@ module IRB
     # Can be either the default <code>IRB.conf[:AUTO_INDENT]</code>, or the
     # mode set by #prompt_mode=
     #
-    # To enable auto-indentation in irb:
+    # To disable auto-indentation in irb:
     #
-    #     IRB.conf[:AUTO_INDENT] = true
+    #     IRB.conf[:AUTO_INDENT] = false
     #
     # or
     #
-    #     irb_context.auto_indent_mode = true
+    #     irb_context.auto_indent_mode = false
     #
     # or
     #
-    #     IRB.CurrentContext.auto_indent_mode = true
+    #     IRB.CurrentContext.auto_indent_mode = false
     #
     # See IRB@Configuration for more information.
     attr_accessor :auto_indent_mode
@@ -206,14 +244,19 @@ module IRB
     #     puts "omg"
     #     # omg
     attr_accessor :echo
+    # Whether to echo for assignment expressions
+    #
+    # Uses IRB.conf[:ECHO_ON_ASSIGNMENT] if available, or defaults to +false+.
+    #
+    #     a = "omg"
+    #     IRB.CurrentContext.echo_on_assignment = true
+    #     a = "omg"
+    #     #=> omg
+    attr_accessor :echo_on_assignment
     # Whether verbose messages are displayed or not.
     #
     # A copy of the default <code>IRB.conf[:VERBOSE]</code>
     attr_accessor :verbose
-    # The debug level of irb
-    #
-    # See #debug_level= for more information.
-    attr_reader :debug_level
 
     # The limit of backtrace lines displayed as top +n+ and tail +n+.
     #
@@ -224,18 +267,33 @@ module IRB
     # See IRB@Command+line+options for more command line options.
     attr_accessor :back_trace_limit
 
-    # Alias for #use_readline
-    alias use_readline? use_readline
+    # Alias for #use_multiline
+    alias use_multiline? use_multiline
+    # Alias for #use_singleline
+    alias use_singleline? use_singleline
+    # backward compatibility
+    alias use_reidline use_multiline
+    # backward compatibility
+    alias use_reidline? use_multiline
+    # backward compatibility
+    alias use_readline use_singleline
+    # backward compatibility
+    alias use_readline? use_singleline
+    # Alias for #use_colorize
+    alias use_colorize? use_colorize
     # Alias for #rc
     alias rc? rc
     alias ignore_sigint? ignore_sigint
     alias ignore_eof? ignore_eof
     alias echo? echo
+    alias echo_on_assignment? echo_on_assignment
 
     # Returns whether messages are displayed or not.
     def verbose?
       if @verbose.nil?
-        if defined?(ReadlineInputMethod) && @io.kind_of?(ReadlineInputMethod)
+        if @io.kind_of?(ReidlineInputMethod)
+          false
+        elsif defined?(ReadlineInputMethod) && @io.kind_of?(ReadlineInputMethod)
           false
         elsif !STDIN.tty? or @io.kind_of?(FileInputMethod)
           true
@@ -248,9 +306,11 @@ module IRB
     end
 
     # Whether #verbose? is +true+, and +input_method+ is either
-    # StdioInputMethod or ReadlineInputMethod, see #io for more information.
+    # StdioInputMethod or ReidlineInputMethod or ReadlineInputMethod, see #io
+    # for more information.
     def prompting?
       verbose? || (STDIN.tty? && @io.kind_of?(StdioInputMethod) ||
+                   @io.kind_of?(ReidlineInputMethod) ||
                    (defined?(ReadlineInputMethod) && @io.kind_of?(ReadlineInputMethod)))
     end
 
@@ -261,7 +321,7 @@ module IRB
     # to #last_value.
     def set_last_value(value)
       @last_value = value
-      @workspace.evaluate self, "_ = IRB.CurrentContext.last_value"
+      @workspace.local_variable_set :_, value
     end
 
     # Sets the +mode+ of the prompt in this context.
@@ -349,34 +409,13 @@ module IRB
       @inspect_mode
     end
 
-    # Obsolete method.
-    #
-    # Can be set using the +--noreadline+ and +--readline+ command line
-    # options.
-    #
-    # See IRB@Command+line+options for more command line options.
-    def use_readline=(opt)
-      print "This method is obsolete."
-      print "Do nothing."
-    end
-
-    # Sets the debug level of irb
-    #
-    # Can also be set using the +--irb_debug+ command line option.
-    #
-    # See IRB@Command+line+options for more command line options.
-    def debug_level=(value)
-      @debug_level = value
-      RubyLex.debug_level = value
-    end
-
-    # Whether or not debug mode is enabled, see #debug_level=.
-    def debug?
-      @debug_level > 0
-    end
-
-    def evaluate(line, line_no) # :nodoc:
+    def evaluate(line, line_no, exception: nil) # :nodoc:
       @line_no = line_no
+      if exception
+        line_no -= 1
+        line = "begin ::Kernel.raise _; rescue _.class\n#{line}\n""end"
+        @workspace.local_variable_set(:_, exception)
+      end
       set_last_value(@workspace.evaluate(self, line, irb_path, line_no))
     end
 
