@@ -2,7 +2,7 @@
 
   math.c -
 
-  $Author: nobu $
+  $Author$
   created at: Tue Jan 25 14:12:56 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -382,8 +382,8 @@ math_atanh(VALUE unused_obj, VALUE x)
     /* check for domain error */
     if (d <  -1.0 || +1.0 <  d) domain_error("atanh");
     /* check for pole error */
-    if (d == -1.0) return DBL2NUM(-INFINITY);
-    if (d == +1.0) return DBL2NUM(+INFINITY);
+    if (d == -1.0) return DBL2NUM(-HUGE_VAL);
+    if (d == +1.0) return DBL2NUM(+HUGE_VAL);
     return DBL2NUM(atanh(d));
 }
 
@@ -426,6 +426,7 @@ math_exp(VALUE unused_obj, VALUE x)
 #endif
 
 static double math_log1(VALUE x);
+FUNC_MINIMIZED(static VALUE math_log(int, const VALUE *, VALUE));
 
 /*
  *  call-seq:
@@ -450,6 +451,12 @@ static double math_log1(VALUE x);
 
 static VALUE
 math_log(int argc, const VALUE *argv, VALUE unused_obj)
+{
+    return rb_math_log(argc, argv);
+}
+
+VALUE
+rb_math_log(int argc, const VALUE *argv)
 {
     VALUE x, base;
     double d;
@@ -488,7 +495,7 @@ math_log1(VALUE x)
     /* check for domain error */
     if (d < 0.0) domain_error("log");
     /* check for pole error */
-    if (d == 0.0) return -INFINITY;
+    if (d == 0.0) return -HUGE_VAL;
 
     return log(d) + numbits * M_LN2; /* log(d * 2 ** numbits) */
 }
@@ -531,7 +538,7 @@ math_log2(VALUE unused_obj, VALUE x)
     /* check for domain error */
     if (d < 0.0) domain_error("log2");
     /* check for pole error */
-    if (d == 0.0) return DBL2NUM(-INFINITY);
+    if (d == 0.0) return DBL2NUM(-HUGE_VAL);
 
     return DBL2NUM(log2(d) + numbits); /* log2(d * 2 ** numbits) */
 }
@@ -561,11 +568,13 @@ math_log10(VALUE unused_obj, VALUE x)
     /* check for domain error */
     if (d < 0.0) domain_error("log10");
     /* check for pole error */
-    if (d == 0.0) return DBL2NUM(-INFINITY);
+    if (d == 0.0) return DBL2NUM(-HUGE_VAL);
 
     return DBL2NUM(log10(d) + numbits * log10(2)); /* log10(d * 2 ** numbits) */
 }
 
+static VALUE rb_math_sqrt(VALUE x);
+
 /*
  *  call-seq:
  *     Math.sqrt(x)    -> Float
@@ -623,7 +632,7 @@ f_signbit(VALUE x)
     return f_negative_p(x);
 }
 
-VALUE
+static VALUE
 rb_math_sqrt(VALUE x)
 {
     double d;
@@ -682,7 +691,14 @@ rb_math_sqrt(VALUE x)
 static VALUE
 math_cbrt(VALUE unused_obj, VALUE x)
 {
-    return DBL2NUM(cbrt(Get_Double(x)));
+    double f = Get_Double(x);
+    double r = cbrt(f);
+#if defined __GLIBC__
+    if (isfinite(r)) {
+	r = (2.0 * r + (f / r / r)) / 3.0;
+    }
+#endif
+    return DBL2NUM(r);
 }
 
 /*
@@ -855,10 +871,10 @@ math_gamma(VALUE unused_obj, VALUE x)
     /* check for domain error */
     if (isinf(d)) {
 	if (signbit(d)) domain_error("gamma");
-	return DBL2NUM(INFINITY);
+	return DBL2NUM(HUGE_VAL);
     }
     if (d == 0.0) {
-	return signbit(d) ? DBL2NUM(-INFINITY) : DBL2NUM(INFINITY);
+	return signbit(d) ? DBL2NUM(-HUGE_VAL) : DBL2NUM(HUGE_VAL);
     }
     if (d == floor(d)) {
 	if (d < 0.0) domain_error("gamma");
@@ -893,11 +909,11 @@ math_lgamma(VALUE unused_obj, VALUE x)
     /* check for domain error */
     if (isinf(d)) {
 	if (signbit(d)) domain_error("lgamma");
-	return rb_assoc_new(DBL2NUM(INFINITY), INT2FIX(1));
+	return rb_assoc_new(DBL2NUM(HUGE_VAL), INT2FIX(1));
     }
     if (d == 0.0) {
 	VALUE vsign = signbit(d) ? INT2FIX(-1) : INT2FIX(+1);
-	return rb_assoc_new(DBL2NUM(INFINITY), vsign);
+	return rb_assoc_new(DBL2NUM(HUGE_VAL), vsign);
     }
     v = DBL2NUM(lgamma_r(d, &sign));
     return rb_assoc_new(v, INT2FIX(sign));
@@ -923,13 +939,6 @@ exp1(cos)
 exp1(cosh)
 exp1(exp)
 exp2(hypot)
-
-VALUE
-rb_math_log(int argc, const VALUE *argv)
-{
-    return math_log(argc, argv, 0);
-}
-
 exp1(sin)
 exp1(sinh)
 #if 0
@@ -975,7 +984,7 @@ InitVM_Math(void)
     rb_define_const(rb_mMath, "PI", DBL2NUM(M_PI));
 
 #ifdef M_E
-    /*  Definition of the mathematical constant E (e) as a Float number. */
+    /*  Definition of the mathematical constant E for Euler's number (e) as a Float number. */
     rb_define_const(rb_mMath, "E", DBL2NUM(M_E));
 #else
     rb_define_const(rb_mMath, "E", DBL2NUM(exp(1.0)));
