@@ -1,12 +1,7 @@
 # frozen_string_literal: true
 
-require "support/helpers"
-require "support/path"
-
 module Spec
   CommandExecution = Struct.new(:command, :working_directory, :exitstatus, :stdout, :stderr) do
-    include RSpec::Matchers::Composable
-
     def to_s
       c = Shellwords.shellsplit(command.strip).map {|s| s.include?("\n") ? " \\\n  <<EOS\n#{s.gsub(/^/, "  ").chomp}\nEOS" : Shellwords.shellescape(s) }
       c = c.reduce("") do |acc, elem|
@@ -27,14 +22,6 @@ module Spec
       @stdboth ||= [stderr, stdout].join("\n").strip
     end
 
-    def bundler_err
-      if Bundler::VERSION.start_with?("1.")
-        stdout
-      else
-        stderr
-      end
-    end
-
     def to_s_verbose
       [
         to_s,
