@@ -1,5 +1,5 @@
-require File.expand_path('../../../spec_helper', __FILE__)
-require File.expand_path('../fixtures/classes', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'fixtures/classes'
 
 describe "Kernel#instance_variables" do
   describe "immediate values" do
@@ -9,7 +9,7 @@ describe "Kernel#instance_variables" do
 
     it "returns the correct array if an instance variable is added" do
       a = 0
-      lambda{ a.instance_variable_set("@test", 1) }.should raise_error(RuntimeError)
+      ->{ a.instance_variable_set("@test", 1) }.should raise_error(RuntimeError)
     end
   end
 
