@@ -1,4 +1,4 @@
-require File.expand_path('../../../spec_helper', __FILE__)
+require_relative '../../spec_helper'
 
 describe "Regexp#===" do
   it "is true if there is a match" do
@@ -17,7 +17,7 @@ describe "Regexp#===" do
     (/a/ === :b).should be_false
   end
 
-  # mirroring https://github.com/ruby/ruby/blob/trunk/test/ruby/test_regexp.rb
+  # mirroring https://github.com/ruby/ruby/blob/master/test/ruby/test_regexp.rb
   it "returns false if the other value cannot be coerced to a string" do
     (/abc/ === nil).should be_false
     (/abc/ === /abc/).should be_false
