@@ -1,5 +1,5 @@
-require File.expand_path('../../spec_helper', __FILE__)
-require File.expand_path('../fixtures/def', __FILE__)
+require_relative '../spec_helper'
+require_relative 'fixtures/def'
 
 # Language-level method behaviour
 describe "Redefining a method" do
@@ -79,6 +79,18 @@ describe "Defining a method" do
   end
 end
 
+describe "An instance method" do
+  it "raises an error with too few arguments" do
+    def foo(a, b); end
+    -> { foo 1 }.should raise_error(ArgumentError, 'wrong number of arguments (given 1, expected 2)')
+  end
+
+  it "raises an error with too many arguments" do
+    def foo(a); end
+    -> { foo 1, 2 }.should raise_error(ArgumentError, 'wrong number of arguments (given 2, expected 1)')
+  end
+end
+
 describe "An instance method definition with a splat" do
   it "accepts an unnamed '*' argument" do
     def foo(*); end;
@@ -101,12 +113,12 @@ describe "An instance method definition with a splat" do
   end
 
   it "allows only a single * argument" do
-    lambda { eval 'def foo(a, *b, *c); end' }.should raise_error(SyntaxError)
+    -> { eval 'def foo(a, *b, *c); end' }.should raise_error(SyntaxError)
   end
 
   it "requires the presence of any arguments that precede the *" do
     def foo(a, b, *c); end
-    lambda { foo 1 }.should raise_error(ArgumentError)
+    -> { foo 1 }.should raise_error(ArgumentError, 'wrong number of arguments (given 1, expected 2+)')
   end
 end
 
@@ -139,7 +151,7 @@ describe "An instance method with a default argument" do
     def foo(a, b = 2)
       [a,b]
     end
-    lambda { foo }.should raise_error(ArgumentError)
+    -> { foo }.should raise_error(ArgumentError, 'wrong number of arguments (given 0, expected 1..2)')
     foo(1).should == [1, 2]
   end
 
@@ -147,7 +159,7 @@ describe "An instance method with a default argument" do
     def foo(a, b = 2, *c)
       [a,b,c]
     end
-    lambda { foo }.should raise_error(ArgumentError)
+    -> { foo }.should raise_error(ArgumentError, 'wrong number of arguments (given 0, expected 1+)')
     foo(1).should == [1,2,[]]
   end
 
@@ -165,17 +177,32 @@ describe "An instance method with a default argument" do
     foo(2,3,3).should == [2,3,[3]]
   end
 
-  it "shadows an existing method with the same name as the local" do
-    def bar
-      1
+  ruby_version_is ''...'2.7' do
+    it "warns and uses a nil value when there is an existing local method with same name" do
+      def bar
+        1
+      end
+      -> {
+        eval "def foo(bar = bar)
+          bar
+        end"
+      }.should complain(/circular argument reference/)
+      foo.should == nil
+      foo(2).should == 2
+    end
+  end
+
+  ruby_version_is '2.7' do
+    it "raises a syntaxError an existing method with the same name as the local variable" do
+      def bar
+        1
+      end
+      -> {
+        eval "def foo(bar = bar)
+          bar
+        end"
+      }.should raise_error(SyntaxError)
     end
-    -> {
-      eval "def foo(bar = bar)
-        bar
-      end"
-    }.should complain(/circular argument reference/)
-    foo.should == nil
-    foo(2).should == 2
   end
 
   it "calls a method with the same name as the local when explicitly using ()" do
@@ -234,10 +261,10 @@ describe "A singleton method definition" do
     (obj==2).should == 2
   end
 
-  it "raises RuntimeError if frozen" do
+  it "raises #{frozen_error_class} if frozen" do
     obj = Object.new
     obj.freeze
-    lambda { def obj.foo; end }.should raise_error(RuntimeError)
+    -> { def obj.foo; end }.should raise_error(frozen_error_class)
   end
 end
 
@@ -310,7 +337,7 @@ describe "A method defined with extreme default arguments" do
   end
 
   it "may use a lambda as a default" do
-    def foo(output = 'a', prc = lambda {|n| output * n})
+    def foo(output = 'a', prc = -> n { output * n })
       prc.call(5)
     end
     foo.should == 'aaaaa'
@@ -356,7 +383,7 @@ describe "A singleton method defined with extreme default arguments" do
 
   it "may use a lambda as a default" do
     a = Object.new
-    def a.foo(output = 'a', prc = lambda {|n| output * n})
+    def a.foo(output = 'a', prc = -> n { output * n })
       prc.call(5)
     end
     a.foo.should == 'aaaaa'
@@ -372,7 +399,7 @@ describe "A method definition inside a metaclass scope" do
     end
 
     DefSpecSingleton.a_class_method.should == DefSpecSingleton
-    lambda { Object.a_class_method }.should raise_error(NoMethodError)
+    -> { Object.a_class_method }.should raise_error(NoMethodError)
   end
 
   it "can create a singleton method" do
@@ -382,15 +409,15 @@ describe "A method definition inside a metaclass scope" do
     end
 
     obj.a_singleton_method.should == obj
-    lambda { Object.new.a_singleton_method }.should raise_error(NoMethodError)
+    -> { Object.new.a_singleton_method }.should raise_error(NoMethodError)
   end
 
-  it "raises RuntimeError if frozen" do
+  it "raises #{frozen_error_class} if frozen" do
     obj = Object.new
     obj.freeze
 
     class << obj
-      lambda { def foo; end }.should raise_error(RuntimeError)
+      -> { def foo; end }.should raise_error(frozen_error_class)
     end
   end
 end
@@ -426,11 +453,11 @@ describe "A nested method definition" do
       end
     end
 
-    lambda { DefSpecNested.a_class_method }.should raise_error(NoMethodError)
+    -> { DefSpecNested.a_class_method }.should raise_error(NoMethodError)
     DefSpecNested.create_class_method.should == DefSpecNested
     DefSpecNested.a_class_method.should == DefSpecNested
-    lambda { Object.a_class_method }.should raise_error(NoMethodError)
-    lambda { DefSpecNested.new.a_class_method }.should raise_error(NoMethodError)
+    -> { Object.a_class_method }.should raise_error(NoMethodError)
+    -> { DefSpecNested.new.a_class_method }.should raise_error(NoMethodError)
   end
 
   it "creates a singleton method when evaluated in the metaclass of an instance" do
@@ -448,7 +475,7 @@ describe "A nested method definition" do
     obj.a_singleton_method.should == obj
 
     other = DefSpecNested.new
-    lambda { other.a_singleton_method }.should raise_error(NoMethodError)
+    -> { other.a_singleton_method }.should raise_error(NoMethodError)
   end
 
   it "creates a method in the surrounding context when evaluated in a def expr.method" do
@@ -488,7 +515,7 @@ describe "A nested method definition" do
     DefSpecNested.should_not have_instance_method :body_method
   end
 
-  it "defines methods as public by default" do
+  it "creates an instance method inside Class.new" do
     cls = Class.new do
       def do_def
         def new_def
@@ -500,6 +527,41 @@ describe "A nested method definition" do
     obj = cls.new
     obj.do_def
     obj.new_def.should == 1
+
+    cls.new.new_def.should == 1
+
+    -> { Object.new.new_def }.should raise_error(NoMethodError)
+  end
+end
+
+describe "A method definition always resets the visibility to public for nested definitions" do
+  it "in Class.new" do
+    cls = Class.new do
+      private
+      def do_def
+        def new_def
+          1
+        end
+      end
+    end
+
+    obj = cls.new
+    -> { obj.do_def }.should raise_error(NoMethodError, /private/)
+    obj.send :do_def
+    obj.new_def.should == 1
+
+    cls.new.new_def.should == 1
+
+    -> { Object.new.new_def }.should raise_error(NoMethodError)
+  end
+
+  it "at the toplevel" do
+    obj = Object.new
+    -> { obj.toplevel_define_other_method }.should raise_error(NoMethodError, /private/)
+    toplevel_define_other_method
+    nested_method_in_toplevel_method.should == 42
+
+    Object.new.nested_method_in_toplevel_method.should == 42
   end
 end
 
@@ -512,7 +574,7 @@ describe "A method definition inside an instance_eval" do
     obj.an_instance_eval_method.should == obj
 
     other = Object.new
-    lambda { other.an_instance_eval_method }.should raise_error(NoMethodError)
+    -> { other.an_instance_eval_method }.should raise_error(NoMethodError)
   end
 
   it "creates a singleton method when evaluated inside a metaclass" do
@@ -525,7 +587,7 @@ describe "A method definition inside an instance_eval" do
     obj.a_metaclass_eval_method.should == obj
 
     other = Object.new
-    lambda { other.a_metaclass_eval_method }.should raise_error(NoMethodError)
+    -> { other.a_metaclass_eval_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method when the receiver is a class" do
@@ -534,7 +596,7 @@ describe "A method definition inside an instance_eval" do
     end
 
     DefSpecNested.an_instance_eval_class_method.should == DefSpecNested
-    lambda { Object.an_instance_eval_class_method }.should raise_error(NoMethodError)
+    -> { Object.an_instance_eval_class_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method when the receiver is an anonymous class" do
@@ -546,7 +608,7 @@ describe "A method definition inside an instance_eval" do
     end
 
     m.klass_method.should == :test
-    lambda { Object.klass_method }.should raise_error(NoMethodError)
+    -> { Object.klass_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method when instance_eval is within class" do
@@ -559,7 +621,7 @@ describe "A method definition inside an instance_eval" do
     end
 
     m.klass_method.should == :test
-    lambda { Object.klass_method }.should raise_error(NoMethodError)
+    -> { Object.klass_method }.should raise_error(NoMethodError)
   end
 end
 
@@ -572,7 +634,7 @@ describe "A method definition inside an instance_exec" do
     end
 
     DefSpecNested.an_instance_exec_class_method.should == 1
-    lambda { Object.an_instance_exec_class_method }.should raise_error(NoMethodError)
+    -> { Object.an_instance_exec_class_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method when the receiver is an anonymous class" do
@@ -586,7 +648,7 @@ describe "A method definition inside an instance_exec" do
     end
 
     m.klass_method.should == 1
-    lambda { Object.klass_method }.should raise_error(NoMethodError)
+    -> { Object.klass_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method when instance_exec is within class" do
@@ -601,7 +663,7 @@ describe "A method definition inside an instance_exec" do
     end
 
     m.klass_method.should == 2
-    lambda { Object.klass_method }.should raise_error(NoMethodError)
+    -> { Object.klass_method }.should raise_error(NoMethodError)
   end
 end
 
@@ -621,7 +683,7 @@ describe "A method definition in an eval" do
     other = DefSpecNested.new
     other.an_eval_instance_method.should == other
 
-    lambda { Object.new.an_eval_instance_method }.should raise_error(NoMethodError)
+    -> { Object.new.an_eval_instance_method }.should raise_error(NoMethodError)
   end
 
   it "creates a class method" do
@@ -637,8 +699,8 @@ describe "A method definition in an eval" do
     DefSpecNestedB.eval_class_method.should == DefSpecNestedB
     DefSpecNestedB.an_eval_class_method.should == DefSpecNestedB
 
-    lambda { Object.an_eval_class_method }.should raise_error(NoMethodError)
-    lambda { DefSpecNestedB.new.an_eval_class_method}.should raise_error(NoMethodError)
+    -> { Object.an_eval_class_method }.should raise_error(NoMethodError)
+    -> { DefSpecNestedB.new.an_eval_class_method}.should raise_error(NoMethodError)
   end
 
   it "creates a singleton method" do
@@ -656,7 +718,7 @@ describe "A method definition in an eval" do
     obj.an_eval_singleton_method.should == obj
 
     other = DefSpecNested.new
-    lambda { other.an_eval_singleton_method }.should raise_error(NoMethodError)
+    -> { other.an_eval_singleton_method }.should raise_error(NoMethodError)
   end
 end
 
@@ -682,7 +744,7 @@ describe "a method definition that sets more than one default parameter all to t
   end
 
   it "only allows overriding the default value of the first such parameter in each set" do
-    lambda { foo(1,2) }.should raise_error(ArgumentError)
+    -> { foo(1,2) }.should raise_error(ArgumentError, 'wrong number of arguments (given 2, expected 0..1)')
   end
 
   def bar(a=b=c=1,d=2)
@@ -693,7 +755,7 @@ describe "a method definition that sets more than one default parameter all to t
     bar.should == [1,1,1,2]
     bar(3).should == [3,nil,nil,2]
     bar(3,4).should == [3,nil,nil,4]
-    lambda { bar(3,4,5) }.should raise_error(ArgumentError)
+    -> { bar(3,4,5) }.should raise_error(ArgumentError, 'wrong number of arguments (given 3, expected 0..2)')
   end
 end
 
@@ -703,7 +765,7 @@ describe "The def keyword" do
       module DefSpecsLambdaVisibility
         private
 
-        lambda {
+        -> {
           def some_method; end
         }.call
       end
