@@ -1,4 +1,5 @@
-require File.expand_path('../../../spec_helper', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'fixtures/set_like'
 require 'set'
 
 describe "Set#proper_subset?" do
@@ -26,9 +27,15 @@ describe "Set#proper_subset?" do
   end
 
   it "raises an ArgumentError when passed a non-Set" do
-    lambda { Set[].proper_subset?([]) }.should raise_error(ArgumentError)
-    lambda { Set[].proper_subset?(1) }.should raise_error(ArgumentError)
-    lambda { Set[].proper_subset?("test") }.should raise_error(ArgumentError)
-    lambda { Set[].proper_subset?(Object.new) }.should raise_error(ArgumentError)
+    -> { Set[].proper_subset?([]) }.should raise_error(ArgumentError)
+    -> { Set[].proper_subset?(1) }.should raise_error(ArgumentError)
+    -> { Set[].proper_subset?("test") }.should raise_error(ArgumentError)
+    -> { Set[].proper_subset?(Object.new) }.should raise_error(ArgumentError)
+  end
+
+  context "when comparing to a Set-like object" do
+    it "returns true if passed a Set-like object that self is a proper subset of" do
+      Set[1, 2, 3].proper_subset?(SetSpecs::SetLike.new([1, 2, 3, 4])).should be_true
+    end
   end
 end
