@@ -1,6 +1,6 @@
-require File.expand_path('../../../spec_helper', __FILE__)
-require File.expand_path('../fixtures/classes', __FILE__)
-require File.expand_path('../shared/each', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'fixtures/classes'
+require_relative 'shared/each'
 
 describe "StringIO#each_line when passed a separator" do
   it_behaves_like :stringio_each_separator, :each_line
@@ -14,8 +14,6 @@ describe "StringIO#each_line when self is not readable" do
   it_behaves_like :stringio_each_not_readable, :each_line
 end
 
-ruby_version_is "2.4" do
-  describe "StringIO#each_line when passed chomp" do
-    it_behaves_like :stringio_each_chomp, :each_line
-  end
+describe "StringIO#each_line when passed chomp" do
+  it_behaves_like :stringio_each_chomp, :each_line
 end
