@@ -2,7 +2,7 @@
 
   struct.c -
 
-  $Author: stomar $
+  $Author$
   created at: Tue Mar 22 18:44:30 JST 1995
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -12,6 +12,7 @@
 #include "internal.h"
 #include "vm_core.h"
 #include "id.h"
+#include "transient_heap.h"
 
 /* only for struct[:field] access */
 enum {
@@ -47,6 +48,12 @@ struct_ivar_get(VALUE c, ID id)
     }
 }
 
+VALUE
+rb_struct_s_keyword_init(VALUE klass)
+{
+    return struct_ivar_get(klass, id_keyword_init);
+}
+
 VALUE
 rb_struct_s_members(VALUE klass)
 {
@@ -132,7 +139,6 @@ static inline int
 struct_member_pos(VALUE s, VALUE name)
 {
     VALUE back = struct_ivar_get(rb_obj_class(s), id_back_members);
-    VALUE const * p;
     long j, mask;
 
     if (UNLIKELY(NIL_P(back))) {
@@ -142,7 +148,6 @@ struct_member_pos(VALUE s, VALUE name)
 	rb_raise(rb_eTypeError, "corrupted struct");
     }
 
-    p = RARRAY_CONST_PTR(back);
     mask = RARRAY_LEN(back);
 
     if (mask <= AREF_HASH_THRESHOLD) {
@@ -152,7 +157,7 @@ struct_member_pos(VALUE s, VALUE name)
 		     mask, RSTRUCT_LEN(s));
 	}
 	for (j = 0; j < mask; j++) {
-	    if (p[j] == name)
+            if (RARRAY_AREF(back, j) == name)
 		return (int)j;
 	}
 	return -1;
@@ -167,9 +172,10 @@ struct_member_pos(VALUE s, VALUE name)
     j = struct_member_pos_ideal(name, mask);
 
     for (;;) {
-	if (p[j] == name)
-	    return FIX2INT(p[j + 1]);
-	if (!RTEST(p[j])) {
+        VALUE e = RARRAY_AREF(back, j);
+        if (e == name)
+            return FIX2INT(RARRAY_AREF(back, j + 1));
+        if (!RTEST(e)) {
 	    return -1;
 	}
 	j = struct_member_pos_probe(j, mask);
@@ -211,7 +217,7 @@ rb_struct_getmember(VALUE obj, ID id)
     }
     rb_name_err_raise("`%1$s' is not a struct member", obj, ID2SYM(id));
 
-    UNREACHABLE;
+    UNREACHABLE_RETURN(Qnil);
 }
 
 static VALUE rb_struct_ref0(VALUE obj) {return RSTRUCT_GET(obj, 0);}
@@ -244,7 +250,6 @@ static void
 rb_struct_modify(VALUE s)
 {
     rb_check_frozen(s);
-    rb_check_trusted(s);
 }
 
 static VALUE
@@ -298,36 +303,44 @@ static VALUE
 rb_struct_s_inspect(VALUE klass)
 {
     VALUE inspect = rb_class_name(klass);
-    if (RTEST(struct_ivar_get(klass, id_keyword_init))) {
+    if (RTEST(rb_struct_s_keyword_init(klass))) {
 	rb_str_cat_cstr(inspect, "(keyword_init: true)");
     }
     return inspect;
 }
 
 static VALUE
-setup_struct(VALUE nstr, VALUE members)
+struct_new_kw(int argc, const VALUE *argv, VALUE klass)
+{
+    return rb_class_new_instance_kw(argc, argv, klass, RB_PASS_CALLED_KEYWORDS);
+}
+
+static VALUE
+setup_struct(VALUE nstr, VALUE members, int keyword_init)
 {
-    const VALUE *ptr_members;
     long i, len;
+    VALUE (*new_func)(int, const VALUE *, VALUE) = rb_class_new_instance;
+
+    if (keyword_init) new_func = struct_new_kw;
 
     members = struct_set_members(nstr, members);
 
     rb_define_alloc_func(nstr, struct_alloc);
-    rb_define_singleton_method(nstr, "new", rb_class_new_instance, -1);
-    rb_define_singleton_method(nstr, "[]", rb_class_new_instance, -1);
+    rb_define_singleton_method(nstr, "new", new_func, -1);
+    rb_define_singleton_method(nstr, "[]", new_func, -1);
     rb_define_singleton_method(nstr, "members", rb_struct_s_members_m, 0);
     rb_define_singleton_method(nstr, "inspect", rb_struct_s_inspect, 0);
-    ptr_members = RARRAY_CONST_PTR(members);
     len = RARRAY_LEN(members);
     for (i=0; i< len; i++) {
-	ID id = SYM2ID(ptr_members[i]);
+        VALUE sym = RARRAY_AREF(members, i);
+        ID id = SYM2ID(sym);
 	VALUE off = LONG2NUM(i);
 
 	if (i < N_REF_FUNC) {
 	    rb_define_method_id(nstr, id, ref_func[i], 0);
 	}
 	else {
-	    define_aref_method(nstr, ptr_members[i], off);
+            define_aref_method(nstr, sym, off);
 	}
 	define_aset_method(nstr, ID2SYM(rb_id_attrset(id)), off);
     }
@@ -429,7 +442,7 @@ rb_struct_define(const char *name, ...)
 
     if (!name) st = anonymous_struct(rb_cStruct);
     else st = new_struct(rb_str_new2(name), rb_cStruct);
-    return setup_struct(st, ary);
+    return setup_struct(st, ary, 0);
 }
 
 VALUE
@@ -442,7 +455,7 @@ rb_struct_define_under(VALUE outer, const char *name, ...)
     ary = struct_make_members_list(ar);
     va_end(ar);
 
-    return setup_struct(rb_define_class_under(outer, name, rb_cStruct), ary);
+    return setup_struct(rb_define_class_under(outer, name, rb_cStruct), ary, 0);
 }
 
 /*
@@ -511,7 +524,7 @@ rb_struct_define_under(VALUE outer, const char *name, ...)
 static VALUE
 rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
 {
-    VALUE name, rest, keyword_init;
+    VALUE name, rest, keyword_init = Qfalse;
     long i;
     VALUE st;
     st_table *tbl;
@@ -527,18 +540,16 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
     }
 
     if (RB_TYPE_P(argv[argc-1], T_HASH)) {
-	VALUE kwargs[1];
 	static ID keyword_ids[1];
 
 	if (!keyword_ids[0]) {
 	    keyword_ids[0] = rb_intern("keyword_init");
 	}
-	rb_get_kwargs(argv[argc-1], keyword_ids, 0, 1, kwargs);
+        rb_get_kwargs(argv[argc-1], keyword_ids, 0, 1, &keyword_init);
+        if (keyword_init == Qundef) {
+            keyword_init = Qfalse;
+        }
 	--argc;
-	keyword_init = kwargs[0];
-    }
-    else {
-	keyword_init = Qfalse;
     }
 
     rest = rb_ident_hash_new();
@@ -546,6 +557,9 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
     tbl = RHASH_TBL(rest);
     for (i=0; i<argc; i++) {
 	VALUE mem = rb_to_symbol(argv[i]);
+        if (rb_is_attrset_sym(mem)) {
+            rb_raise(rb_eArgError, "invalid struct member: %"PRIsVALUE, mem);
+        }
 	if (st_insert(tbl, mem, Qtrue)) {
 	    rb_raise(rb_eArgError, "duplicate member: %"PRIsVALUE, mem);
 	}
@@ -560,10 +574,10 @@ rb_struct_s_def(int argc, VALUE *argv, VALUE klass)
     else {
 	st = new_struct(name, klass);
     }
-    setup_struct(st, rest);
+    setup_struct(st, rest, (int)keyword_init);
     rb_ivar_set(st, id_keyword_init, keyword_init);
     if (rb_block_given_p()) {
-	rb_mod_module_eval(0, 0, st);
+        rb_mod_module_eval(0, 0, st);
     }
 
     return st;
@@ -616,9 +630,9 @@ rb_struct_initialize_m(int argc, const VALUE *argv, VALUE self)
 
     rb_struct_modify(self);
     n = num_members(klass);
-    if (argc > 0 && RTEST(struct_ivar_get(klass, id_keyword_init))) {
+    if (argc > 0 && RTEST(rb_struct_s_keyword_init(klass))) {
 	struct struct_hash_set_arg arg;
-	if (argc > 2 || !RB_TYPE_P(argv[0], T_HASH)) {
+	if (argc > 1 || !RB_TYPE_P(argv[0], T_HASH)) {
 	    rb_raise(rb_eArgError, "wrong number of arguments (given %d, expected 0)", argc);
 	}
 	rb_mem_clear((VALUE *)RSTRUCT_CONST_PTR(self), n);
@@ -650,6 +664,43 @@ rb_struct_initialize(VALUE self, VALUE values)
     return rb_struct_initialize_m(RARRAY_LENINT(values), RARRAY_CONST_PTR(values), self);
 }
 
+static VALUE *
+struct_heap_alloc(VALUE st, size_t len)
+{
+    VALUE *ptr = rb_transient_heap_alloc((VALUE)st, sizeof(VALUE) * len);
+
+    if (ptr) {
+        RSTRUCT_TRANSIENT_SET(st);
+        return ptr;
+    }
+    else {
+        RSTRUCT_TRANSIENT_UNSET(st);
+        return ALLOC_N(VALUE, len);
+    }
+}
+
+#if USE_TRANSIENT_HEAP
+void
+rb_struct_transient_heap_evacuate(VALUE obj, int promote)
+{
+    if (RSTRUCT_TRANSIENT_P(obj)) {
+        const VALUE *old_ptr = rb_struct_const_heap_ptr(obj);
+        VALUE *new_ptr;
+        long len = RSTRUCT_LEN(obj);
+
+        if (promote) {
+            new_ptr = ALLOC_N(VALUE, len);
+            FL_UNSET_RAW(obj, RSTRUCT_TRANSIENT_FLAG);
+        }
+        else {
+            new_ptr = struct_heap_alloc(obj, len);
+        }
+        MEMCPY(new_ptr, old_ptr, VALUE, len);
+        RSTRUCT(obj)->as.heap.ptr = new_ptr;
+    }
+}
+#endif
+
 static VALUE
 struct_alloc(VALUE klass)
 {
@@ -664,9 +715,9 @@ struct_alloc(VALUE klass)
 	rb_mem_clear((VALUE *)st->as.ary, n);
     }
     else {
-	st->as.heap.ptr = ALLOC_N(VALUE, n);
-	rb_mem_clear((VALUE *)st->as.heap.ptr, n);
-	st->as.heap.len = n;
+        st->as.heap.ptr = struct_heap_alloc((VALUE)st, n);
+        rb_mem_clear((VALUE *)st->as.heap.ptr, n);
+        st->as.heap.len = n;
     }
 
     return (VALUE)st;
@@ -820,7 +871,6 @@ inspect_struct(VALUE s, VALUE dummy, int recur)
 	rb_str_append(str, rb_inspect(RSTRUCT_GET(s, i)));
     }
     rb_str_cat2(str, ">");
-    OBJ_INFECT(str, s);
 
     return str;
 }
@@ -859,13 +909,19 @@ rb_struct_to_a(VALUE s)
 
 /*
  *  call-seq:
- *     struct.to_h     -> hash
+ *     struct.to_h                        -> hash
+ *     struct.to_h {|name, value| block } -> hash
  *
  *  Returns a Hash containing the names and values for the struct's members.
  *
+ *  If a block is given, the results of the block on each pair of the receiver
+ *  will be used as pairs.
+ *
  *     Customer = Struct.new(:name, :address, :zip)
  *     joe = Customer.new("Joe Smith", "123 Maple, Anytown NC", 12345)
  *     joe.to_h[:address]   #=> "123 Maple, Anytown NC"
+ *     joe.to_h{|name, value| [name.upcase, value.to_s.upcase]}[:ADDRESS]
+ *                          #=> "123 MAPLE, ANYTOWN NC"
  */
 
 static VALUE
@@ -874,9 +930,44 @@ rb_struct_to_h(VALUE s)
     VALUE h = rb_hash_new_with_size(RSTRUCT_LEN(s));
     VALUE members = rb_struct_members(s);
     long i;
+    int block_given = rb_block_given_p();
 
     for (i=0; i<RSTRUCT_LEN(s); i++) {
-	rb_hash_aset(h, rb_ary_entry(members, i), RSTRUCT_GET(s, i));
+        VALUE k = rb_ary_entry(members, i), v = RSTRUCT_GET(s, i);
+        if (block_given)
+            rb_hash_set_pair(h, rb_yield_values(2, k, v));
+        else
+            rb_hash_aset(h, k, v);
+    }
+    return h;
+}
+
+static VALUE
+rb_struct_deconstruct_keys(VALUE s, VALUE keys)
+{
+    VALUE h;
+    long i;
+
+    if (NIL_P(keys)) {
+        return rb_struct_to_h(s);
+    }
+    if (UNLIKELY(!RB_TYPE_P(keys, T_ARRAY))) {
+	rb_raise(rb_eTypeError,
+                 "wrong argument type %"PRIsVALUE" (expected Array or nil)",
+                 rb_obj_class(keys));
+
+    }
+    if (RSTRUCT_LEN(s) < RARRAY_LEN(keys)) {
+        return rb_hash_new_with_size(0);
+    }
+    h = rb_hash_new_with_size(RARRAY_LEN(keys));
+    for (i=0; i<RARRAY_LEN(keys); i++) {
+        VALUE key = RARRAY_AREF(keys, i);
+        int i = rb_struct_pos(s, &key);
+        if (i < 0) {
+            return h;
+        }
+        rb_hash_aset(h, key, RSTRUCT_GET(s, i));
     }
     return h;
 }
@@ -1053,6 +1144,8 @@ rb_struct_values_at(int argc, VALUE *argv, VALUE s)
  *  call-seq:
  *     struct.select {|obj| block }  -> array
  *     struct.select                 -> enumerator
+ *     struct.filter {|obj| block }  -> array
+ *     struct.filter                 -> enumerator
  *
  *  Yields each member value from the struct to the block and returns an Array
  *  containing the member values from the +struct+ for which the given block
@@ -1061,6 +1154,8 @@ rb_struct_values_at(int argc, VALUE *argv, VALUE s)
  *     Lots = Struct.new(:a, :b, :c, :d, :e, :f)
  *     l = Lots.new(11, 22, 33, 44, 55, 66)
  *     l.select {|v| v.even? }   #=> [22, 44, 66]
+ *
+ *  Struct#filter is an alias for Struct#select.
  */
 
 static VALUE
@@ -1084,15 +1179,12 @@ rb_struct_select(int argc, VALUE *argv, VALUE s)
 static VALUE
 recursive_equal(VALUE s, VALUE s2, int recur)
 {
-    const VALUE *ptr, *ptr2;
     long i, len;
 
     if (recur) return Qtrue; /* Subtle! */
-    ptr = RSTRUCT_CONST_PTR(s);
-    ptr2 = RSTRUCT_CONST_PTR(s2);
     len = RSTRUCT_LEN(s);
     for (i=0; i<len; i++) {
-	if (!rb_equal(ptr[i], ptr2[i])) return Qfalse;
+        if (!rb_equal(RSTRUCT_GET(s, i), RSTRUCT_GET(s2, i))) return Qfalse;
     }
     return Qtrue;
 }
@@ -1140,31 +1232,26 @@ rb_struct_hash(VALUE s)
     long i, len;
     st_index_t h;
     VALUE n;
-    const VALUE *ptr;
 
     h = rb_hash_start(rb_hash(rb_obj_class(s)));
-    ptr = RSTRUCT_CONST_PTR(s);
     len = RSTRUCT_LEN(s);
     for (i = 0; i < len; i++) {
-	n = rb_hash(ptr[i]);
+        n = rb_hash(RSTRUCT_GET(s, i));
 	h = rb_hash_uint(h, NUM2LONG(n));
     }
     h = rb_hash_end(h);
-    return INT2FIX(h);
+    return ST2FIX(h);
 }
 
 static VALUE
 recursive_eql(VALUE s, VALUE s2, int recur)
 {
-    const VALUE *ptr, *ptr2;
     long i, len;
 
     if (recur) return Qtrue; /* Subtle! */
-    ptr = RSTRUCT_CONST_PTR(s);
-    ptr2 = RSTRUCT_CONST_PTR(s2);
     len = RSTRUCT_LEN(s);
     for (i=0; i<len; i++) {
-	if (!rb_eql(ptr[i], ptr2[i])) return Qfalse;
+        if (!rb_eql(RSTRUCT_GET(s, i), RSTRUCT_GET(s2, i))) return Qfalse;
     }
     return Qtrue;
 }
@@ -1291,10 +1378,14 @@ InitVM_Struct(void)
     rb_define_method(rb_cStruct, "[]", rb_struct_aref, 1);
     rb_define_method(rb_cStruct, "[]=", rb_struct_aset, 2);
     rb_define_method(rb_cStruct, "select", rb_struct_select, -1);
+    rb_define_method(rb_cStruct, "filter", rb_struct_select, -1);
     rb_define_method(rb_cStruct, "values_at", rb_struct_values_at, -1);
 
     rb_define_method(rb_cStruct, "members", rb_struct_members_m, 0);
     rb_define_method(rb_cStruct, "dig", rb_struct_dig, -1);
+
+    rb_define_method(rb_cStruct, "deconstruct", rb_struct_to_a, 0);
+    rb_define_method(rb_cStruct, "deconstruct_keys", rb_struct_deconstruct_keys, 1);
 }
 
 #undef rb_intern
