@@ -58,6 +58,13 @@ static const uint32_t RB_ID_SERIAL_MAX = /* 256M on LP32 */
     ((sizeof(ID)-sizeof(rb_id_serial_t))*CHAR_BIT < RUBY_ID_SCOPE_SHIFT ?
      RUBY_ID_SCOPE_SHIFT : 0);
 
+typedef struct {
+    rb_id_serial_t last_id;
+    st_table *str_sym;
+    VALUE ids;
+    VALUE dsymbol_fstr_hash;
+} rb_symbols_t;
+
 static inline rb_id_serial_t
 rb_id_to_serial(ID id)
 {
@@ -102,7 +109,7 @@ is_global_name_punct(const int c)
     return (ruby_global_name_punct_bits[(c - 0x20) / 32] >> (c % 32)) & 1;
 }
 
-ID rb_intern_cstr_without_pindown(const char *, long, rb_encoding *);
+int rb_enc_symname_type(const char *name, long len, rb_encoding *enc, unsigned int allowed_attrset);
 
 RUBY_SYMBOL_EXPORT_BEGIN
 
