@@ -1,4 +1,4 @@
-# frozen_string_literal: false
+# frozen_string_literal: true
 require 'test/unit'
 require 'cgi'
 require 'stringio'
@@ -15,7 +15,7 @@ class CGIUtilTest < Test::Unit::TestCase
       'REQUEST_METHOD' => 'GET',
       'SCRIPT_NAME' => nil,
     )
-    @str1="&<>\" \xE3\x82\x86\xE3\x82\x93\xE3\x82\x86\xE3\x82\x93"
+    @str1="&<>\" \xE3\x82\x86\xE3\x82\x93\xE3\x82\x86\xE3\x82\x93".dup
     @str1.force_encoding("UTF-8") if defined?(::Encoding)
   end
 
@@ -25,32 +25,41 @@ class CGIUtilTest < Test::Unit::TestCase
 
 
   def test_cgi_escape
-    assert_equal('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93', CGI::escape(@str1))
-    assert_equal('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93'.ascii_only?, CGI::escape(@str1).ascii_only?) if defined?(::Encoding)
+    assert_equal('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93', CGI.escape(@str1))
+    assert_equal('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93'.ascii_only?, CGI.escape(@str1).ascii_only?) if defined?(::Encoding)
+  end
+
+  def test_cgi_escape_with_unreserved_characters
+    assert_equal("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~",
+                 CGI.escape("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"),
+                 "should not escape any unreserved characters, as per RFC3986 Section 2.3")
   end
 
   def test_cgi_escape_with_invalid_byte_sequence
     assert_nothing_raised(ArgumentError) do
-      assert_equal('%C0%3C%3C', CGI::escape("\xC0\<\<".force_encoding("UTF-8")))
+      assert_equal('%C0%3C%3C', CGI.escape("\xC0\<\<".dup.force_encoding("UTF-8")))
     end
   end
 
   def test_cgi_escape_preserve_encoding
-    assert_equal(Encoding::US_ASCII, CGI::escape("\xC0\<\<".force_encoding("US-ASCII")).encoding)
-    assert_equal(Encoding::ASCII_8BIT, CGI::escape("\xC0\<\<".force_encoding("ASCII-8BIT")).encoding)
-    assert_equal(Encoding::UTF_8, CGI::escape("\xC0\<\<".force_encoding("UTF-8")).encoding)
+    assert_equal(Encoding::US_ASCII, CGI.escape("\xC0\<\<".dup.force_encoding("US-ASCII")).encoding)
+    assert_equal(Encoding::ASCII_8BIT, CGI.escape("\xC0\<\<".dup.force_encoding("ASCII-8BIT")).encoding)
+    assert_equal(Encoding::UTF_8, CGI.escape("\xC0\<\<".dup.force_encoding("UTF-8")).encoding)
   end
 
   def test_cgi_unescape
-    assert_equal(@str1, CGI::unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93'))
-    assert_equal(@str1.encoding, CGI::unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93').encoding) if defined?(::Encoding)
+    str = CGI.unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93')
+    assert_equal(@str1, str)
+    return unless defined?(::Encoding)
+
+    assert_equal(@str1.encoding, str.encoding)
     assert_equal("\u{30E1 30E2 30EA 691C 7D22}", CGI.unescape("\u{30E1 30E2 30EA}%E6%A4%9C%E7%B4%A2"))
   end
 
   def test_cgi_unescape_preserve_encoding
-    assert_equal(Encoding::US_ASCII, CGI::unescape("%C0%3C%3C".force_encoding("US-ASCII")).encoding)
-    assert_equal(Encoding::ASCII_8BIT, CGI::unescape("%C0%3C%3C".force_encoding("ASCII-8BIT")).encoding)
-    assert_equal(Encoding::UTF_8, CGI::unescape("%C0%3C%3C".force_encoding("UTF-8")).encoding)
+    assert_equal(Encoding::US_ASCII, CGI.unescape("%C0%3C%3C".dup.force_encoding("US-ASCII")).encoding)
+    assert_equal(Encoding::ASCII_8BIT, CGI.unescape("%C0%3C%3C".dup.force_encoding("ASCII-8BIT")).encoding)
+    assert_equal(Encoding::UTF_8, CGI.unescape("%C0%3C%3C".dup.force_encoding("UTF-8")).encoding)
   end
 
   def test_cgi_unescape_accept_charset
@@ -64,23 +73,23 @@ class CGIUtilTest < Test::Unit::TestCase
   end
 
   def test_cgi_pretty
-    assert_equal("<HTML>\n  <BODY>\n  </BODY>\n</HTML>\n",CGI::pretty("<HTML><BODY></BODY></HTML>"))
-    assert_equal("<HTML>\n\t<BODY>\n\t</BODY>\n</HTML>\n",CGI::pretty("<HTML><BODY></BODY></HTML>","\t"))
+    assert_equal("<HTML>\n  <BODY>\n  </BODY>\n</HTML>\n",CGI.pretty("<HTML><BODY></BODY></HTML>"))
+    assert_equal("<HTML>\n\t<BODY>\n\t</BODY>\n</HTML>\n",CGI.pretty("<HTML><BODY></BODY></HTML>","\t"))
   end
 
   def test_cgi_escapeHTML
-    assert_equal("&#39;&amp;&quot;&gt;&lt;", CGI::escapeHTML("'&\"><"))
+    assert_equal("&#39;&amp;&quot;&gt;&lt;", CGI.escapeHTML("'&\"><"))
   end
 
   def test_cgi_escape_html_duplicated
-    orig = "Ruby".force_encoding("US-ASCII")
-    str = CGI::escapeHTML(orig)
+    orig = "Ruby".dup.force_encoding("US-ASCII")
+    str = CGI.escapeHTML(orig)
     assert_equal(orig, str)
     assert_not_same(orig, str)
   end
 
   def assert_cgi_escape_html_preserve_encoding(str, encoding)
-    assert_equal(encoding, CGI::escapeHTML(str.dup.force_encoding(encoding)).encoding)
+    assert_equal(encoding, CGI.escapeHTML(str.dup.force_encoding(encoding)).encoding)
   end
 
   def test_cgi_escape_html_preserve_encoding
@@ -90,26 +99,19 @@ class CGIUtilTest < Test::Unit::TestCase
     end
   end
 
-  def test_cgi_escape_html_preserve_tainted
-    assert_not_predicate CGI::escapeHTML("'&\"><"),       :tainted?
-    assert_predicate     CGI::escapeHTML("'&\"><".taint), :tainted?
-    assert_not_predicate CGI::escapeHTML("Ruby"),         :tainted?
-    assert_predicate     CGI::escapeHTML("Ruby".taint),   :tainted?
-  end
-
   def test_cgi_escape_html_dont_freeze
-    assert_not_predicate CGI::escapeHTML("'&\"><".dup),    :frozen?
-    assert_not_predicate CGI::escapeHTML("'&\"><".freeze), :frozen?
-    assert_not_predicate CGI::escapeHTML("Ruby".dup),      :frozen?
-    assert_not_predicate CGI::escapeHTML("Ruby".freeze),   :frozen?
+    assert_not_predicate CGI.escapeHTML("'&\"><".dup),    :frozen?
+    assert_not_predicate CGI.escapeHTML("'&\"><".freeze), :frozen?
+    assert_not_predicate CGI.escapeHTML("Ruby".dup),      :frozen?
+    assert_not_predicate CGI.escapeHTML("Ruby".freeze),   :frozen?
   end
 
   def test_cgi_unescapeHTML
-    assert_equal("'&\"><", CGI::unescapeHTML("&#39;&amp;&quot;&gt;&lt;"))
+    assert_equal("'&\"><", CGI.unescapeHTML("&#39;&amp;&quot;&gt;&lt;"))
   end
 
   def test_cgi_unescapeHTML_invalid
-    assert_equal('&<&amp>&quot&abcdefghijklmn', CGI::unescapeHTML('&&lt;&amp&gt;&quot&abcdefghijklmn'))
+    assert_equal('&<&amp>&quot&abcdefghijklmn', CGI.unescapeHTML('&&lt;&amp&gt;&quot&abcdefghijklmn'))
   end
 
   Encoding.list.each do |enc|
@@ -120,10 +122,10 @@ class CGIUtilTest < Test::Unit::TestCase
       next
     else
       define_method("test_cgi_escapeHTML:#{enc.name}") do
-        assert_equal(escaped, CGI::escapeHTML(unescaped))
+        assert_equal(escaped, CGI.escapeHTML(unescaped))
       end
       define_method("test_cgi_unescapeHTML:#{enc.name}") do
-        assert_equal(unescaped, CGI::unescapeHTML(escaped))
+        assert_equal(unescaped, CGI.unescapeHTML(escaped))
       end
     end
   end
@@ -137,16 +139,16 @@ class CGIUtilTest < Test::Unit::TestCase
       next
     else
       define_method("test_cgi_escape:#{enc.name}") do
-        assert_equal(escaped, CGI::escape(unescaped))
+        assert_equal(escaped, CGI.escape(unescaped))
       end
       define_method("test_cgi_unescape:#{enc.name}") do
-        assert_equal(unescaped, CGI::unescape(escaped, enc))
+        assert_equal(unescaped, CGI.unescape(escaped, enc))
       end
     end
   end
 
   def test_cgi_unescapeHTML_uppercasecharacter
-    assert_equal("\xE3\x81\x82\xE3\x81\x84\xE3\x81\x86", CGI::unescapeHTML("&#x3042;&#x3044;&#X3046;"))
+    assert_equal("\xE3\x81\x82\xE3\x81\x84\xE3\x81\x86", CGI.unescapeHTML("&#x3042;&#x3044;&#X3046;"))
   end
 
   def test_cgi_include_escape
@@ -162,8 +164,11 @@ class CGIUtilTest < Test::Unit::TestCase
   end
 
   def test_cgi_include_unescape
-    assert_equal(@str1, unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93'))
-    assert_equal(@str1.encoding, unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93').encoding) if defined?(::Encoding)
+    str = unescape('%26%3C%3E%22+%E3%82%86%E3%82%93%E3%82%86%E3%82%93')
+    assert_equal(@str1, str)
+    return unless defined?(::Encoding)
+
+    assert_equal(@str1.encoding, str.encoding)
     assert_equal("\u{30E1 30E2 30EA 691C 7D22}", unescape("\u{30E1 30E2 30EA}%E6%A4%9C%E7%B4%A2"))
   end
 
