@@ -7,29 +7,14 @@ class TestDRbCore < Test::Unit::TestCase
   include DRbCore
 
   def setup
+    super
     setup_service 'ut_drb.rb'
-    super
-  end
-
-  def teardown
-    super
-    DRbService.finish
   end
 end
 
-class TestDRbYield < Test::Unit::TestCase
+module DRbYield
   include DRbBase
 
-  def setup
-    setup_service 'ut_drb.rb'
-    super
-  end
-
-  def teardown
-    super
-    DRbService.finish
-  end
-
   def test_01_one
     @there.echo_yield_1([]) {|one|
       assert_equal([], one)
@@ -118,18 +103,25 @@ class TestDRbYield < Test::Unit::TestCase
     @there.xarray_each {|x| assert_kind_of(XArray, x)}
     @there.xarray_each {|*x| assert_kind_of(XArray, x[0])}
   end
+end
 
-  def test_06_taint
-    x = proc {}
-    assert_not_predicate(x, :tainted?)
-    @there.echo_yield(x) {|o|
-      assert_equal(x, o)
-      assert_not_predicate(x, :tainted?)
-    }
+class TestDRbYield < Test::Unit::TestCase
+  include DRbYield
+
+  def setup
+    super
+    setup_service 'ut_drb.rb'
   end
 end
 
-class TestDRbRubyYield < TestDRbYield
+class TestDRbRubyYield < Test::Unit::TestCase
+  include DRbYield
+
+  def setup
+    @there = self
+    super
+  end
+
   def echo_yield(*arg)
     yield(*arg)
   end
@@ -153,15 +145,11 @@ class TestDRbRubyYield < TestDRbYield
     end
   end
 
-  def setup
-    @there = self
-  end
-
-  def teardown
-  end
 end
 
-class TestDRbRuby18Yield < TestDRbRubyYield
+class TestDRbRuby18Yield < Test::Unit::TestCase
+  include DRbYield
+
   class YieldTest18
     def echo_yield(*arg, &proc)
       proc.call(*arg)
@@ -188,6 +176,7 @@ class TestDRbRuby18Yield < TestDRbRubyYield
 
   def setup
     @there = YieldTest18.new
+    super
   end
 end
 
@@ -195,13 +184,8 @@ class TestDRbAry < Test::Unit::TestCase
   include DRbAry
 
   def setup
+    super
     setup_service 'ut_array.rb'
-    super
-  end
-
-  def teardown
-    super
-    DRbService.finish
   end
 end
 
@@ -209,8 +193,8 @@ class TestDRbMServer < Test::Unit::TestCase
   include DRbBase
 
   def setup
-    setup_service 'ut_drb.rb'
     super
+    setup_service 'ut_drb.rb'
     @server = (1..3).collect do |n|
       DRb::DRbServer.new("druby://localhost:0", Onecky.new(n.to_s))
     end
@@ -221,7 +205,6 @@ class TestDRbMServer < Test::Unit::TestCase
       s.stop_service
     end
     super
-    DRbService.finish
   end
 
   def test_01
@@ -229,14 +212,11 @@ class TestDRbMServer < Test::Unit::TestCase
   end
 end
 
-class TestDRbSafe1 < TestDRbAry
+class TestDRbSafe1 < Test::Unit::TestCase
+  include DRbAry
   def setup
-    setup_service 'ut_safe1.rb'
-  end
-
-  def teardown
     super
-    DRbService.finish
+    setup_service 'ut_safe1.rb'
   end
 end
 
@@ -244,13 +224,8 @@ class TestDRbLarge < Test::Unit::TestCase
   include DRbBase
 
   def setup
+    super
     setup_service 'ut_large.rb'
-    super
-  end
-
-  def teardown
-    super
-    DRbService.finish
   end
 
   def test_01_large_ary
@@ -333,13 +308,8 @@ class TestBug4409 < Test::Unit::TestCase
   include DRbBase
 
   def setup
+    super
     setup_service 'ut_eq.rb'
-    super
-  end
-
-  def teardown
-    super
-    DRbService.finish
   end
 
   def test_bug4409
@@ -348,4 +318,35 @@ class TestBug4409 < Test::Unit::TestCase
   end
 end
 
+class TestDRbAnyToS < Test::Unit::TestCase
+  class BO < BasicObject
+  end
+
+  def test_any_to_s
+    server = DRb::DRbServer.new('druby://:0')
+    server.singleton_class.send(:public, :any_to_s)
+    assert_equal("foo:String", server.any_to_s("foo"))
+    assert_match(/\A#<DRbTests::TestDRbAnyToS::BO:0x[0-9a-f]+>\z/, server.any_to_s(BO.new))
+    server.stop_service
+    server.thread.join
+    DRb::DRbConn.stop_pool
+  end
+end
+
+class TestDRbTCP < Test::Unit::TestCase
+  def test_immediate_close
+    server = DRb::DRbServer.new('druby://:0')
+    host, port, = DRb::DRbTCPSocket.send(:parse_uri, server.uri)
+    socket = TCPSocket.open host, port
+    socket.shutdown
+    socket.close
+    client = DRb::DRbTCPSocket.new(server.uri, socket)
+    assert client
+    client.close
+    server.stop_service
+    server.thread.join
+    DRb::DRbConn.stop_pool
+  end
+end
+
 end
