@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require_relative 'helper'
 
 module DTrace
@@ -13,11 +14,11 @@ module DTrace
     end
 
     def test_many_lit
-      trap_probe(probe, '[1,2,3,4]') { |_,rbfile,saw|
-        saw = saw.map(&:split).find_all { |num, file, line|
+      trap_probe(probe, '[1,2,3,4]') { |_,rbfile,orig|
+        saw = orig.map(&:split).find_all { |num, file, line|
           file == rbfile && num == '4' && line == '1'
         }
-        assert_operator saw.length, :>, 0
+        assert_operator saw.length, :>, 0, orig
       }
     end
 
@@ -25,7 +26,7 @@ module DTrace
     def probe type = 'array'
       <<-eoprobe
 ruby$target:::#{type}-create
-/arg1/
+/arg1 && arg2/
 {
   printf("%d %s %d\\n", arg0, copyinstr(arg1), arg2);
 }
