@@ -1,9 +1,11 @@
+# frozen_string_literal: false
 require "monitor"
-require "thread"
 
 require "test/unit"
 
 class TestMonitor < Test::Unit::TestCase
+  Queue = Thread::Queue
+
   def setup
     @monitor = Monitor.new
   end
@@ -33,6 +35,45 @@ class TestMonitor < Test::Unit::TestCase
     assert_equal((1..10).to_a, ary)
   end
 
+  def test_exit
+    m = Monitor.new
+    m.enter
+    assert_equal true, m.mon_owned?
+    m.exit
+    assert_equal false, m.mon_owned?
+
+    assert_raise ThreadError do
+      m.exit
+    end
+
+    assert_equal false, m.mon_owned?
+
+    m.enter
+    Thread.new{
+      assert_raise(ThreadError) do
+        m.exit
+      end
+    }.join
+    assert_equal true, m.mon_owned?
+    m.exit
+  end
+
+  def test_enter_second_after_killed_thread
+    th = Thread.start {
+      @monitor.enter
+      Thread.current.kill
+      @monitor.exit
+    }
+    th.join
+    @monitor.enter
+    @monitor.exit
+    th2 = Thread.start {
+      @monitor.enter
+      @monitor.exit
+    }
+    assert_join_threads([th, th2])
+  end
+
   def test_synchronize
     ary = []
     queue = Queue.new
@@ -111,6 +152,53 @@ class TestMonitor < Test::Unit::TestCase
     assert_join_threads([th, th2])
   end
 
+  def test_try_enter_second_after_killed_thread
+    th = Thread.start {
+      assert_equal(true, @monitor.try_enter)
+      Thread.current.kill
+      @monitor.exit
+    }
+    th.join
+    assert_equal(true, @monitor.try_enter)
+    @monitor.exit
+    th2 = Thread.start {
+      assert_equal(true, @monitor.try_enter)
+      @monitor.exit
+    }
+    assert_join_threads([th, th2])
+  end
+
+  def test_mon_locked_and_owned
+    queue1 = Queue.new
+    queue2 = Queue.new
+    th = Thread.start {
+      @monitor.enter
+      queue1.enq(nil)
+      queue2.deq
+      @monitor.exit
+      queue1.enq(nil)
+    }
+    queue1.deq
+    assert(@monitor.mon_locked?)
+    assert(!@monitor.mon_owned?)
+
+    queue2.enq(nil)
+    queue1.deq
+    assert(!@monitor.mon_locked?)
+
+    @monitor.enter
+    assert @monitor.mon_locked?
+    assert @monitor.mon_owned?
+    @monitor.exit
+
+    @monitor.synchronize do
+      assert @monitor.mon_locked?
+      assert @monitor.mon_owned?
+    end
+  ensure
+    th.join
+  end
+
   def test_cond
     cond = @monitor.new_cond
 
@@ -135,6 +223,19 @@ class TestMonitor < Test::Unit::TestCase
     assert_join_threads([th, th2])
   end
 
+  class NewCondTest
+    include MonitorMixin
+    attr_reader :cond
+    def initialize
+      @cond = new_cond
+      super # mon_initialize
+    end
+  end
+
+  def test_new_cond_before_initialize
+    assert NewCondTest.new.cond.instance_variable_get(:@monitor) != nil
+  end
+
   def test_timedwait
     cond = @monitor.new_cond
     b = "foo"
@@ -206,4 +307,26 @@ class TestMonitor < Test::Unit::TestCase
 #     end
 #     cumber_thread.kill
   end
+
+  def test_wait_interruption
+    cond = @monitor.new_cond
+
+    th = Thread.start {
+      @monitor.synchronize do
+        begin
+          cond.wait(0.1)
+          @monitor.mon_owned?
+        rescue Interrupt
+          @monitor.mon_owned?
+        end
+      end
+    }
+    sleep(0.1)
+    th.raise(Interrupt)
+
+    begin
+      assert_equal true, th.value
+    rescue Interrupt
+    end
+  end
 end
