@@ -5,8 +5,14 @@ require 'stringio'
 require 'tempfile'
 
 class TestPsych < Psych::TestCase
+
+  def setup
+    @orig_verbose, $VERBOSE = $VERBOSE, nil
+  end
+
   def teardown
     Psych.domain_types.clear
+    $VERBOSE = @orig_verbose
   end
 
   def test_line_width_invalid
@@ -60,6 +66,22 @@ class TestPsych < Psych::TestCase
     end
   end
 
+  def test_parse
+    assert_equal %w[a b], Psych.parse("- a\n- b").to_ruby
+  end
+
+  def test_parse_default_fallback
+    assert_equal false, Psych.parse("")
+  end
+
+  def test_parse_raises_on_bad_input
+    assert_raises(Psych::SyntaxError) { Psych.parse("--- `") }
+  end
+
+  def test_parse_with_fallback
+    assert_equal 42, Psych.parse("", fallback: 42)
+  end
+
   def test_non_existing_class_on_deserialize
     e = assert_raises(ArgumentError) do
       Psych.load("--- !ruby/object:NonExistent\nfoo: 1")
@@ -103,9 +125,44 @@ class TestPsych < Psych::TestCase
     assert_equal %w{ foo bar }, docs
   end
 
+  def test_load_stream_default_fallback
+    assert_equal [], Psych.load_stream("")
+  end
+
+  def test_load_stream_raises_on_bad_input
+    assert_raises(Psych::SyntaxError) { Psych.load_stream("--- `") }
+  end
+
   def test_parse_stream
     docs = Psych.parse_stream("--- foo\n...\n--- bar\n...")
-    assert_equal %w{ foo bar }, docs.children.map { |x| x.transform }
+    assert_equal(%w[foo bar], docs.children.map(&:transform))
+  end
+
+  def test_parse_stream_with_block
+    docs = []
+    Psych.parse_stream("--- foo\n...\n--- bar\n...") do |node|
+      docs << node
+    end
+
+    assert_equal %w[foo bar], docs.map(&:to_ruby)
+  end
+
+  def test_parse_stream_default_fallback
+    docs = Psych.parse_stream("")
+    assert_equal [], docs.children.map(&:to_ruby)
+  end
+
+  def test_parse_stream_with_block_default_fallback
+    docs = []
+    Psych.parse_stream("") do |node|
+      docs << node
+    end
+
+    assert_equal [], docs.map(&:to_ruby)
+  end
+
+  def test_parse_stream_raises_on_bad_input
+    assert_raises(Psych::SyntaxError) { Psych.parse_stream("--- `") }
   end
 
   def test_add_builtin_type
@@ -121,20 +178,45 @@ class TestPsych < Psych::TestCase
 
   def test_domain_types
     got = nil
-    Psych.add_domain_type 'foo.bar/2002', 'foo' do |type, val|
+    Psych.add_domain_type 'foo.bar,2002', 'foo' do |type, val|
       got = val
     end
 
-    Psych.load('--- !foo.bar/2002:foo hello')
+    Psych.load('--- !foo.bar,2002/foo hello')
     assert_equal 'hello', got
 
-    Psych.load("--- !foo.bar/2002:foo\n- hello\n- world")
+    Psych.load("--- !foo.bar,2002/foo\n- hello\n- world")
     assert_equal %w{ hello world }, got
 
-    Psych.load("--- !foo.bar/2002:foo\nhello: world")
+    Psych.load("--- !foo.bar,2002/foo\nhello: world")
     assert_equal({ 'hello' => 'world' }, got)
   end
 
+  def test_load_default_fallback
+    assert_equal false, Psych.load("")
+  end
+
+  def test_load_with_fallback
+    assert_equal 42, Psych.load("", "file", fallback: 42)
+  end
+
+  def test_load_with_fallback_nil_or_false
+    assert_nil Psych.load("", "file", fallback: nil)
+    assert_equal false, Psych.load("", "file", fallback: false)
+  end
+
+  def test_load_with_fallback_hash
+    assert_equal Hash.new, Psych.load("", "file", fallback: Hash.new)
+  end
+
+  def test_load_with_fallback_for_nil
+    assert_nil Psych.load("--- null", "file", fallback: 42)
+  end
+
+  def test_load_with_fallback_for_false
+    assert_equal false, Psych.load("--- false", "file", fallback: 42)
+  end
+
   def test_load_file
     Tempfile.create(['yikes', 'yml']) {|t|
       t.binmode
@@ -144,12 +226,49 @@ class TestPsych < Psych::TestCase
     }
   end
 
+  def test_load_file_default_fallback
+    Tempfile.create(['empty', 'yml']) {|t|
+      assert_equal false, Psych.load_file(t.path)
+    }
+  end
+
   def test_load_file_with_fallback
+    Tempfile.create(['empty', 'yml']) {|t|
+      assert_equal 42, Psych.load_file(t.path, fallback: 42)
+    }
+  end
+
+  def test_load_file_with_fallback_nil_or_false
+    Tempfile.create(['empty', 'yml']) {|t|
+      assert_nil Psych.load_file(t.path, fallback: nil)
+      assert_equal false, Psych.load_file(t.path, fallback: false)
+    }
+  end
+
+  def test_load_file_with_fallback_hash
     Tempfile.create(['empty', 'yml']) {|t|
       assert_equal Hash.new, Psych.load_file(t.path, fallback: Hash.new)
     }
   end
 
+  def test_load_file_with_fallback_for_nil
+    Tempfile.create(['nil', 'yml']) {|t|
+      t.binmode
+      t.write('--- null')
+      t.close
+      assert_nil Psych.load_file(t.path, fallback: 42)
+    }
+  end
+
+  def test_load_file_with_fallback_for_false
+    Tempfile.create(['false', 'yml']) {|t|
+      t.binmode
+      t.write('--- false')
+      t.close
+      assert_equal false, Psych.load_file(t.path, fallback: 42)
+    }
+  end
+
   def test_parse_file
     Tempfile.create(['yikes', 'yml']) {|t|
       t.binmode
@@ -159,6 +278,12 @@ class TestPsych < Psych::TestCase
     }
   end
 
+  def test_parse_file_default_fallback
+    Tempfile.create(['empty', 'yml']) do |t|
+      assert_equal false, Psych.parse_file(t.path)
+    end
+  end
+
   def test_degenerate_strings
     assert_equal false, Psych.load('    ')
     assert_equal false, Psych.parse('   ')
@@ -170,13 +295,16 @@ class TestPsych < Psych::TestCase
     types = []
     appender = lambda { |*args| types << args }
 
-    Psych.add_domain_type('example.com:2002', 'foo', &appender)
+    Psych.add_builtin_type('foo', &appender)
+    Psych.add_domain_type('example.com,2002', 'foo', &appender)
     Psych.load <<-eoyml
-- !tag:example.com:2002:foo bar
+- !tag:yaml.org,2002:foo bar
+- !tag:example.com,2002:foo bar
     eoyml
 
     assert_equal [
-      ["tag:example.com:2002:foo", "bar"]
+      ["tag:yaml.org,2002:foo", "bar"],
+      ["tag:example.com,2002:foo", "bar"]
     ], types
   end
 
