@@ -1,7 +1,7 @@
 # coding: UTF-8
 # frozen_string_literal: true
 
-require 'rdoc/test_case'
+require_relative 'helper'
 require 'rdoc/markup/block_quote'
 require 'rdoc/markdown'
 
@@ -717,7 +717,7 @@ Some text. ^[With a footnote]
   def test_parse_note_no_notes
     @parser.notes = false
 
-    assert_raises RDoc::Markdown::ParseError do
+    assert_raise RDoc::Markdown::ParseError do
       parse "Some text.[^1]"
     end
   end
