@@ -428,6 +428,7 @@ class TestPack < Test::Unit::TestCase
     assert_operator(4, :<=, [1].pack("L!").bytesize)
   end
 
+  require 'rbconfig'
   def test_pack_unpack_qQ
     s1 = [578437695752307201, -506097522914230529].pack("q*")
     s2 = [578437695752307201, 17940646550795321087].pack("Q*")
@@ -437,6 +438,7 @@ class TestPack < Test::Unit::TestCase
 
     # Note: q! and Q! should not work on platform which has no long long type.
     # Is there a such platform now?
+    # @shyouhei: Yes. gcc -ansi is one of such platform.
     s1 = [578437695752307201, -506097522914230529].pack("q!*")
     s2 = [578437695752307201, 17940646550795321087].pack("Q!*")
     assert_equal([578437695752307201, -506097522914230529], s2.unpack("q!*"))
@@ -446,7 +448,7 @@ class TestPack < Test::Unit::TestCase
     assert_equal(8, [1].pack("Q").bytesize)
     assert_operator(8, :<=, [1].pack("q!").bytesize)
     assert_operator(8, :<=, [1].pack("Q!").bytesize)
-  end
+  end if RbConfig::CONFIG['HAVE_LONG_LONG']
 
   def test_pack_unpack_jJ
     # Note: we assume that the size of intptr_t and uintptr_t equals to the size
@@ -549,7 +551,7 @@ class TestPack < Test::Unit::TestCase
     assert_equal([nil], "\x00".unpack("@1C")) # is it OK?
     assert_raise(ArgumentError) { "\x00".unpack("@2C") }
 
-    pos = (1 << [nil].pack("p").bytesize * 8) - 100 # -100
+    pos = RbConfig::LIMITS["UINTPTR_MAX"] - 99 # -100
     assert_raise(RangeError) {"0123456789".unpack("@#{pos}C10")}
   end
 
@@ -691,7 +693,7 @@ EXPECTED
     assert_equal(["pre===31after"], "pre===31after".unpack("M"))
 
     bug = '[ruby-core:83055] [Bug #13949]'
-    s = "abcdef".unpack("M").first
+    s = "abcdef".unpack1("M")
     assert_equal(Encoding::ASCII_8BIT, s.encoding)
     assert_predicate(s, :ascii_only?, bug)
   end
@@ -798,9 +800,14 @@ EXPECTED
       [].pack("\x7f")
     }
     assert_warning(/\A(.* in '\u{3042}'\n)+\z/) {
-      EnvUtil.with_default_external(Encoding::UTF_8) {
-        [].pack("\u{3042}")
-      }
+      [].pack("\u{3042}")
+    }
+
+    assert_warning(/\A.* in '.*U'\Z/) {
+      assert_equal "\000", [0].pack("\0U")
+    }
+    assert_warning(/\A.* in '.*U'\Z/) {
+      "\000".unpack("\0U")
     }
   end
 
@@ -823,10 +830,43 @@ EXPECTED
     end;
   end
 
+  def test_pack_with_buffer
+    buf = String.new(capacity: 100)
+
+    assert_raise_with_message(FrozenError, /frozen/) {
+      [0xDEAD_BEEF].pack('N', buffer: 'foo'.freeze)
+    }
+    assert_raise_with_message(TypeError, /must be String/) {
+      [0xDEAD_BEEF].pack('N', buffer: Object.new)
+    }
+
+    addr = [buf].pack('p')
+
+    [0xDEAD_BEEF].pack('N', buffer: buf)
+    assert_equal "\xDE\xAD\xBE\xEF", buf
+
+    [0xBABE_F00D].pack('@4N', buffer: buf)
+    assert_equal "\xDE\xAD\xBE\xEF\xBA\xBE\xF0\x0D", buf
+    assert_equal addr, [buf].pack('p')
+
+    [0xBAAD_FACE].pack('@10N', buffer: buf)
+    assert_equal "\xDE\xAD\xBE\xEF\xBA\xBE\xF0\x0D\0\0\xBA\xAD\xFA\xCE", buf
+
+    assert_equal addr, [buf].pack('p')
+  end
+
   def test_unpack_with_block
     ret = []; "ABCD".unpack("CCCC") {|v| ret << v }
     assert_equal [65, 66, 67, 68], ret
-    ret = []; "A".unpack("B*") {|v| ret << v }
+    ret = []; "A".unpack("B*") {|v| ret << v.dup }
     assert_equal ["01000001"], ret
   end
+
+  def test_unpack1
+    assert_equal 65, "A".unpack1("C")
+    assert_equal 68, "ABCD".unpack1("x3C")
+    assert_equal 0x3042, "\u{3042 3044 3046}".unpack1("U*")
+    assert_equal "hogefuga", "aG9nZWZ1Z2E=".unpack1("m")
+    assert_equal "01000001", "A".unpack1("B*")
+  end
 end
