@@ -4,7 +4,7 @@ require 'test/unit'
 class TestVMDump < Test::Unit::TestCase
   def assert_darwin_vm_dump_works(args)
     skip if RUBY_PLATFORM !~ /darwin/
-    assert_in_out_err(args, "", [], [:*, /^.* main \+ \d+$/, :*, /^\[IMPORTANT\]/, :*])
+    assert_in_out_err(args, "", [], /^\[IMPORTANT\]/)
   end
 
   def test_darwin_invalid_call
