@@ -2,7 +2,7 @@
 # frozen_string_literal: true
 
 require 'rubygems/package/tar_test_case'
-require 'rubygems/simple_gem'
+require 'digest'
 
 class TestGemPackage < Gem::Package::TarTestCase
 
@@ -24,6 +24,8 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_class_new_old_format
+    skip "jruby can't require the simple_gem file" if Gem.java_platform?
+    require_relative "simple_gem"
     File.open 'old_format.gem', 'wb' do |io|
       io.write SIMPLE_GEM
     end
@@ -105,6 +107,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_build_time_source_date_epoch
+    epoch = ENV["SOURCE_DATE_EPOCH"]
     ENV["SOURCE_DATE_EPOCH"] = "123456789"
 
     spec = Gem::Specification.new 'build', '1'
@@ -114,10 +117,37 @@ class TestGemPackage < Gem::Package::TarTestCase
     spec.date = Time.at 0
     spec.rubygems_version = Gem::Version.new '0'
 
-
     package = Gem::Package.new spec.file_name
 
     assert_equal Time.at(ENV["SOURCE_DATE_EPOCH"].to_i).utc, package.build_time
+  ensure
+    ENV["SOURCE_DATE_EPOCH"] = epoch
+  end
+
+  def test_build_time_source_date_epoch_automatically_set
+    epoch = ENV["SOURCE_DATE_EPOCH"]
+    ENV["SOURCE_DATE_EPOCH"] = nil
+
+    start_time = Time.now.utc.to_i
+
+    spec = Gem::Specification.new 'build', '1'
+    spec.summary = 'build'
+    spec.authors = 'build'
+    spec.files = ['lib/code.rb']
+    spec.rubygems_version = Gem::Version.new '0'
+
+    package = Gem::Package.new spec.file_name
+
+    end_time = Time.now.utc.to_i
+
+    assert_kind_of Time, package.build_time
+
+    build_time = package.build_time.to_i
+
+    assert_operator(start_time, :<=, build_time)
+    assert_operator(build_time, :<=, end_time)
+  ensure
+    ENV["SOURCE_DATE_EPOCH"] = epoch
   end
 
   def test_add_files
@@ -126,8 +156,13 @@ class TestGemPackage < Gem::Package::TarTestCase
 
     FileUtils.mkdir_p 'lib/empty'
 
-    File.open 'lib/code.rb',  'w' do |io| io.write '# lib/code.rb'  end
-    File.open 'lib/extra.rb', 'w' do |io| io.write '# lib/extra.rb' end
+    File.open 'lib/code.rb',  'w' do |io|
+      io.write '# lib/code.rb'
+    end
+
+    File.open 'lib/extra.rb', 'w' do |io|
+      io.write '# lib/extra.rb'
+    end
 
     package = Gem::Package.new 'bogus.gem'
     package.spec = spec
@@ -154,7 +189,10 @@ class TestGemPackage < Gem::Package::TarTestCase
     spec.files = %w[lib/code.rb lib/code_sym.rb lib/code_sym2.rb]
 
     FileUtils.mkdir_p 'lib'
-    File.open 'lib/code.rb',  'w' do |io| io.write '# lib/code.rb'  end
+
+    File.open 'lib/code.rb',  'w' do |io|
+      io.write '# lib/code.rb'
+    end
 
     # NOTE: 'code.rb' is correct, because it's relative to lib/code_sym.rb
     begin
@@ -409,6 +447,33 @@ class TestGemPackage < Gem::Package::TarTestCase
     assert_equal %w[lib/code.rb], reader.contents
   end
 
+  def test_raw_spec
+    data_tgz = util_tar_gz { }
+
+    gem = util_tar do |tar|
+      tar.add_file 'data.tar.gz', 0644 do |io|
+        io.write data_tgz.string
+      end
+
+      tar.add_file 'metadata.gz', 0644 do |io|
+        Zlib::GzipWriter.wrap io do |gzio|
+          gzio.write @spec.to_yaml
+        end
+      end
+    end
+
+    gem_path = "#{@destination}/test.gem"
+
+    File.open gem_path, "wb" do |io|
+      io.write gem.string
+    end
+
+    spec, metadata = Gem::Package.raw_spec(gem_path)
+
+    assert_equal @spec, spec
+    assert_match @spec.to_yaml, metadata.force_encoding("UTF-8")
+  end
+
   def test_contents
     package = Gem::Package.new @gem
 
@@ -430,7 +495,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_extract_files_empty
-    data_tgz = util_tar_gz do end
+    data_tgz = util_tar_gz { }
 
     gem = util_tar do |tar|
       tar.add_file 'data.tar.gz', 0644 do |io|
@@ -459,7 +524,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file '/absolute.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file '/absolute.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     e = assert_raises Gem::Package::PathError do
@@ -474,7 +541,10 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file    'relative.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file    'relative.rb', 0644 do |io|
+        io.write 'hi'
+      end
+
       tar.mkdir       'lib',         0755
       tar.add_symlink 'lib/foo.rb', '../relative.rb', 0644
     end
@@ -503,7 +573,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     tgz_io = util_tar_gz do |tar|
       tar.mkdir       'lib',               0755
       tar.add_symlink 'lib/link', '../..', 0644
-      tar.add_file    'lib/link/outside.txt', 0644 do |io| io.write 'hi' end
+      tar.add_file    'lib/link/outside.txt', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     # Extract into a subdirectory of @destination; if this test fails it writes
@@ -526,12 +598,48 @@ class TestGemPackage < Gem::Package::TarTestCase
     end
   end
 
+  def test_extract_symlink_parent_doesnt_delete_user_dir
+    package = Gem::Package.new @gem
+
+    # Extract into a subdirectory of @destination; if this test fails it writes
+    # a file outside destination_subdir, but we want the file to remain inside
+    # @destination so it will be cleaned up.
+    destination_subdir = File.join @destination, 'subdir'
+    FileUtils.mkdir_p destination_subdir
+
+    destination_user_dir = File.join @destination, 'user'
+    destination_user_subdir = File.join destination_user_dir, 'dir'
+    FileUtils.mkdir_p destination_user_subdir
+
+    tgz_io = util_tar_gz do |tar|
+      tar.add_symlink 'link', destination_user_dir, 16877
+      tar.add_symlink 'link/dir', '.', 16877
+    end
+
+    e = assert_raises(Gem::Package::PathError, Errno::EACCES) do
+      package.extract_tar_gz tgz_io, destination_subdir
+    end
+
+    assert_path_exists destination_user_subdir
+
+    if Gem::Package::PathError === e
+      assert_equal("installing into parent path #{destination_user_subdir} of " +
+                  "#{destination_subdir} is not allowed", e.message)
+    elsif win_platform?
+      skip "symlink - must be admin with no UAC on Windows"
+    else
+      raise e
+    end
+  end
+
   def test_extract_tar_gz_directory
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
       tar.mkdir    'lib',        0755
-      tar.add_file 'lib/foo.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file 'lib/foo.rb', 0644 do |io|
+        io.write 'hi'
+      end
       tar.mkdir    'lib/foo',    0755
     end
 
@@ -548,7 +656,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file './dot_slash.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file './dot_slash.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     package.extract_tar_gz tgz_io, @destination
@@ -561,7 +671,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     tgz_io = util_tar_gz do |tar|
-      tar.add_file '.dot_file.rb', 0644 do |io| io.write 'hi' end
+      tar.add_file '.dot_file.rb', 0644 do |io|
+        io.write 'hi'
+      end
     end
 
     package.extract_tar_gz tgz_io, @destination
@@ -575,7 +687,9 @@ class TestGemPackage < Gem::Package::TarTestCase
       package = Gem::Package.new @gem
 
       tgz_io = util_tar_gz do |tar|
-        tar.add_file 'foo/file.rb', 0644 do |io| io.write 'hi' end
+        tar.add_file 'foo/file.rb', 0644 do |io|
+          io.write 'hi'
+        end
       end
 
       package.extract_tar_gz tgz_io, @destination.upcase
@@ -589,12 +703,12 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     file = 'file.rb'.dup
-    file.taint
+    file.taint if RUBY_VERSION < '2.7'
 
     destination = package.install_location file, @destination
 
     assert_equal File.join(@destination, 'file.rb'), destination
-    refute destination.tainted?
+    refute destination.tainted? if RUBY_VERSION < '2.7'
   end
 
   def test_install_location_absolute
@@ -628,14 +742,14 @@ class TestGemPackage < Gem::Package::TarTestCase
     package = Gem::Package.new @gem
 
     file = 'foo//file.rb'.dup
-    file.taint
+    file.taint if RUBY_VERSION < '2.7'
 
     destination = @destination.sub '/', '//'
 
     destination = package.install_location file, destination
 
     assert_equal File.join(@destination, 'foo', 'file.rb'), destination
-    refute destination.tainted?
+    refute destination.tainted? if RUBY_VERSION < '2.7'
   end
 
   def test_install_location_relative
@@ -780,6 +894,7 @@ class TestGemPackage < Gem::Package::TarTestCase
   end
 
   def test_verify_corrupt
+    skip "jruby strips the null byte and does not think it's corrupt" if Gem.java_platform?
     tf = Tempfile.open 'corrupt' do |io|
       data = Gem::Util.gzip 'a' * 10
       io.write \
@@ -1005,6 +1120,11 @@ class TestGemPackage < Gem::Package::TarTestCase
     assert_equal @spec, package.spec
   end
 
+  def test_gem_attr
+    package = Gem::Package.new(@gem)
+    assert_equal(@gem, package.gem.path)
+  end
+
   def test_spec_from_io
     # This functionality is used by rubygems.org to extract spec data from an
     # uploaded gem before it is written to storage.
@@ -1040,7 +1160,9 @@ class TestGemPackage < Gem::Package::TarTestCase
     tgz_io = StringIO.new
 
     # can't wrap TarWriter because it seeks
-    Zlib::GzipWriter.wrap tgz_io do |io| io.write tar_io.string end
+    Zlib::GzipWriter.wrap tgz_io do |io|
+      io.write tar_io.string
+    end
 
     StringIO.new tgz_io.string
   end
