@@ -3,6 +3,7 @@ require 'rubygems/test_case'
 require 'rubygems/request_set'
 
 class TestGemRequestSet < Gem::TestCase
+
   def setup
     super
 
@@ -52,7 +53,7 @@ class TestGemRequestSet < Gem::TestCase
     rs = Gem::RequestSet.new
     installed = []
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts 'gem "a"'
       io.flush
 
@@ -78,7 +79,7 @@ class TestGemRequestSet < Gem::TestCase
 
     rs = Gem::RequestSet.new
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts 'gem "a"'
       io.flush
 
@@ -104,7 +105,7 @@ Gems to install:
     rs = Gem::RequestSet.new
     installed = []
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts 'gem "a"'
     end
 
@@ -128,7 +129,7 @@ Gems to install:
 
     rs = Gem::RequestSet.new
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts 'gem "a"'
       io.flush
 
@@ -150,7 +151,7 @@ Gems to install:
     rs = Gem::RequestSet.new
     installed = []
 
-    open 'gem.deps.rb.lock', 'w' do |io|
+    File.open 'gem.deps.rb.lock', 'w' do |io|
       io.puts <<-LOCKFILE
 GEM
   remote: #{@gem_repo}
@@ -167,7 +168,7 @@ DEPENDENCIES
       LOCKFILE
     end
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts 'gem "b"'
     end
 
@@ -190,7 +191,7 @@ DEPENDENCIES
     rs = Gem::RequestSet.new
     installed = []
 
-    open 'gem.deps.rb', 'w' do |io|
+    File.open 'gem.deps.rb', 'w' do |io|
       io.puts <<-GEM_DEPS
 gem "a"
 ruby "0"
@@ -218,7 +219,7 @@ ruby "0"
       assert_kind_of Gem::RequestSet::GemDependencyAPI, gem_deps
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
 
     assert_equal [dep('a')], rs.dependencies
 
@@ -239,7 +240,7 @@ ruby "0"
       assert_kind_of Gem::RequestSet::GemDependencyAPI, gem_deps
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
 
     assert_equal [dep('a')], rs.dependencies
   end
@@ -254,7 +255,7 @@ ruby "0"
       rs.load_gemdeps io.path, [:test]
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
 
     assert_empty rs.dependencies
   end
@@ -311,8 +312,14 @@ ruby "0"
   end
 
   def test_resolve_development_shallow
-    a = util_spec 'a', 1 do |s| s.add_development_dependency 'b' end
-    b = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    a = util_spec 'a', 1 do |s|
+      s.add_development_dependency 'b'
+    end
+
+    b = util_spec 'b', 1 do |s|
+      s.add_development_dependency 'c'
+    end
+
     c = util_spec 'c', 1
 
     a_spec = Gem::Resolver::SpecSpecification.new nil, a
@@ -346,7 +353,7 @@ ruby "0"
       rs.load_gemdeps io.path
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
 
     res = rs.resolve
     assert_equal 1, res.size
@@ -410,7 +417,7 @@ ruby "0"
       rs.load_gemdeps io.path
       io
     end
-    tf.close! if tf.respond_to? :close!
+    tf.close!
 
     res = rs.resolve
     assert_equal 2, res.size
@@ -528,8 +535,14 @@ ruby "0"
   end
 
   def test_sorted_requests_development_shallow
-    a = util_spec 'a', 1 do |s| s.add_development_dependency 'b' end
-    b = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    a = util_spec 'a', 1 do |s|
+      s.add_development_dependency 'b'
+    end
+
+    b = util_spec 'b', 1 do |s|
+      s.add_development_dependency 'c'
+    end
+
     c = util_spec 'c', 1
 
     rs = Gem::RequestSet.new
@@ -547,8 +560,14 @@ ruby "0"
   end
 
   def test_tsort_each_child_development
-    a = util_spec 'a', 1 do |s| s.add_development_dependency 'b' end
-    b = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    a = util_spec 'a', 1 do |s|
+      s.add_development_dependency 'b'
+    end
+
+    b = util_spec 'b', 1 do |s|
+      s.add_development_dependency 'c'
+    end
+
     c = util_spec 'c', 1
 
     rs = Gem::RequestSet.new
@@ -570,8 +589,14 @@ ruby "0"
   end
 
   def test_tsort_each_child_development_shallow
-    a = util_spec 'a', 1 do |s| s.add_development_dependency 'b' end
-    b = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    a = util_spec 'a', 1 do |s|
+      s.add_development_dependency 'b'
+    end
+
+    b = util_spec 'b', 1 do |s|
+      s.add_development_dependency 'c'
+    end
+
     c = util_spec 'c', 1
 
     rs = Gem::RequestSet.new
