@@ -1,4 +1,4 @@
-# frozen_string_literal: false
+# frozen_string_literal: true
 require 'rubygems/test_case'
 require 'rubygems/source'
 
@@ -208,16 +208,16 @@ class TestGemSourceGit < Gem::TestCase
     installed = Gem::Source::Installed.new
     vendor    = Gem::Source::Vendor.new 'vendor/foo'
 
-    assert_equal( 0, git.      <=>(git),       'git    <=> git')
+    assert_equal(0, git.      <=>(git),       'git    <=> git')
 
-    assert_equal( 1, git.      <=>(remote),    'git    <=> remote')
+    assert_equal(1, git.      <=>(remote),    'git    <=> remote')
     assert_equal(-1, remote.   <=>(git),       'remote <=> git')
 
-    assert_equal( 1, git.      <=>(installed), 'git       <=> installed')
+    assert_equal(1, git.      <=>(installed), 'git       <=> installed')
     assert_equal(-1, installed.<=>(git),       'installed <=> git')
 
     assert_equal(-1, git.      <=>(vendor),    'git       <=> vendor')
-    assert_equal( 1, vendor.   <=>(git),       'vendor    <=> git')
+    assert_equal(1, vendor.   <=>(git),       'vendor    <=> git')
   end
 
   def test_specs
@@ -229,18 +229,13 @@ class TestGemSourceGit < Gem::TestCase
       Dir.chdir 'b' do
         b = Gem::Specification.new 'b', 1
 
-        open 'b.gemspec', 'w' do |io|
+        File.open 'b.gemspec', 'w' do |io|
           io.write b.to_ruby
         end
 
         system @git, 'add', 'b.gemspec'
         system @git, 'commit', '--quiet', '-m', 'add b/b.gemspec'
       end
-
-      FileUtils.touch 'c.gemspec'
-
-      system @git, 'add', 'c.gemspec'
-      system @git, 'commit', '--quiet', '-m', 'add c.gemspec'
     end
 
     specs = nil
@@ -306,4 +301,3 @@ class TestGemSourceGit < Gem::TestCase
   end
 
 end
-
