@@ -22,13 +22,14 @@ class TestStringIO < Test::Unit::TestCase
     assert_raise(ArgumentError) { StringIO.new('', 'rx') }
     assert_raise(ArgumentError) { StringIO.new('', 'rbt') }
     assert_raise(TypeError) { StringIO.new(nil) }
-    assert_raise(TypeError) { StringIO.new('str', nil) }
 
     o = Object.new
     def o.to_str
       nil
     end
     assert_raise(TypeError) { StringIO.new(o) }
+
+    o = Object.new
     def o.to_str
       'str'
     end
@@ -79,6 +80,8 @@ class TestStringIO < Test::Unit::TestCase
     assert_equal("def\n", stringio.gets(""))
     assert_raise(TypeError){StringIO.new("").gets(1, 1)}
     assert_nothing_raised {StringIO.new("").gets(nil, nil)}
+
+    assert_string("", Encoding::UTF_8, StringIO.new("foo").gets(0))
   end
 
   def test_gets_chomp
@@ -94,6 +97,8 @@ class TestStringIO < Test::Unit::TestCase
     stringio = StringIO.new("abc\n\ndef\n")
     assert_equal("abc\n", stringio.gets("", chomp: true))
     assert_equal("def", stringio.gets("", chomp: true))
+
+    assert_string("", Encoding::UTF_8, StringIO.new("\n").gets(chomp: true))
   end
 
   def test_gets_chomp_eol
@@ -175,17 +180,6 @@ class TestStringIO < Test::Unit::TestCase
     f.close unless f.closed?
   end
 
-  def test_write_infection
-    bug9769 = '[ruby-dev:48118] [Bug #9769]'
-    s = "".untaint
-    f = StringIO.new(s, "w")
-    f.print("bar".taint)
-    f.close
-    assert_predicate(s, :tainted?, bug9769)
-  ensure
-    f.close unless f.closed?
-  end
-
   def test_write_encoding
     s = "".force_encoding(Encoding::UTF_8)
     f = StringIO.new(s)
@@ -194,14 +188,23 @@ class TestStringIO < Test::Unit::TestCase
   end
 
   def test_write_integer_overflow
-    long_max = (1 << (RbConfig::SIZEOF["long"] * 8 - 1)) - 1
     f = StringIO.new
-    f.pos = long_max
+    f.pos = RbConfig::LIMITS["LONG_MAX"]
     assert_raise(ArgumentError) {
       f.write("pos + len overflows")
     }
   end
 
+  def test_write_with_multiple_arguments
+    s = ""
+    f = StringIO.new(s, "w")
+    f.write("foo", "bar")
+    f.close
+    assert_equal("foobar", s)
+  ensure
+    f.close unless f.closed?
+  end
+
   def test_set_encoding
     bug10285 = '[ruby-core:65240] [Bug #10285]'
     f = StringIO.new()
@@ -438,6 +441,10 @@ class TestStringIO < Test::Unit::TestCase
     t.ungetbyte("\u{30eb 30d3 30fc}")
     assert_equal(0, t.pos)
     assert_equal("\u{30eb 30d3 30fc}\u7d05\u7389bar\n", s)
+
+    assert_nothing_raised {t.ungetbyte(-1)}
+    assert_nothing_raised {t.ungetbyte(256)}
+    assert_nothing_raised {t.ungetbyte(1<<64)}
   end
 
   def test_ungetc
@@ -453,6 +460,9 @@ class TestStringIO < Test::Unit::TestCase
     f.ungetc("y".ord)
     assert_equal("y", f.getc)
     assert_equal("2", f.getc)
+
+    assert_raise(RangeError) {f.ungetc(0x1ffffff)}
+    assert_raise(RangeError) {f.ungetc(0xffffffffffffff)}
   ensure
     f.close unless f.closed?
   end
@@ -567,13 +577,20 @@ class TestStringIO < Test::Unit::TestCase
     assert_equal("\u3042\u3044", f.read(nil, nil), bug5207)
     f.rewind
     s = ""
-    f.read(nil, s)
+    assert_same(s, f.read(nil, s))
     assert_equal("\u3042\u3044", s, bug5207)
     f.rewind
     # not empty buffer
     s = "0123456789"
-    f.read(nil, s)
+    assert_same(s, f.read(nil, s))
     assert_equal("\u3042\u3044", s)
+
+    bug13806 = '[ruby-core:82349] [Bug #13806]'
+    assert_string("", Encoding::UTF_8, f.read, bug13806)
+    assert_string("", Encoding::UTF_8, f.read(nil, nil), bug13806)
+    s.force_encoding(Encoding::US_ASCII)
+    assert_same(s, f.read(nil, s))
+    assert_string("", Encoding::UTF_8, s, bug13806)
   end
 
   def test_readpartial
@@ -698,9 +715,10 @@ class TestStringIO < Test::Unit::TestCase
     s = StringIO.new
     s.freeze
     bug = '[ruby-core:33648]'
-    assert_raise(RuntimeError, bug) {s.puts("foo")}
-    assert_raise(RuntimeError, bug) {s.string = "foo"}
-    assert_raise(RuntimeError, bug) {s.reopen("")}
+    exception_class = defined?(FrozenError) ? FrozenError : RuntimeError
+    assert_raise(exception_class, bug) {s.puts("foo")}
+    assert_raise(exception_class, bug) {s.string = "foo"}
+    assert_raise(exception_class, bug) {s.reopen("")}
   end
 
   def test_frozen_string
@@ -737,7 +755,7 @@ class TestStringIO < Test::Unit::TestCase
 
   def test_overflow
     skip if RbConfig::SIZEOF["void*"] > RbConfig::SIZEOF["long"]
-    limit = (1 << (RbConfig::SIZEOF["void*"]*8-1)) - 0x10
+    limit = RbConfig::LIMITS["INTPTR_MAX"] - 0x10
     assert_separately(%w[-rstringio], "#{<<-"begin;"}\n#{<<-"end;"}")
     begin;
       limit = #{limit}
@@ -753,4 +771,33 @@ class TestStringIO < Test::Unit::TestCase
       assert_equal(0x100000, s.pos)
     end;
   end
+
+  def test_encoding_write
+    s = StringIO.new("", "w:utf-32be")
+    s.print "abc"
+    assert_equal("abc".encode("utf-32be"), s.string)
+  end
+
+  def test_encoding_read
+    s = StringIO.new("abc".encode("utf-32be"), "r:utf-8")
+    assert_equal("\0\0\0a\0\0\0b\0\0\0c", s.read)
+  end
+
+  %w/UTF-8 UTF-16BE UTF-16LE UTF-32BE UTF-32LE/.each do |name|
+    define_method("test_strip_bom:#{name}") do
+      text = "\uFEFF\u0100a"
+      content = text.encode(name)
+      result = StringIO.new(content, mode: 'rb:BOM|UTF-8').read
+      assert_equal(Encoding.find(name), result.encoding, name)
+      assert_equal(content[1..-1].b, result.b, name)
+
+      StringIO.open(content) {|f|
+        assert_equal(Encoding.find(name), f.set_encoding_by_bom)
+      }
+    end
+  end
+
+  def assert_string(content, encoding, str, mesg = nil)
+    assert_equal([content, encoding], [str, str.encoding], mesg)
+  end
 end
