@@ -22,6 +22,13 @@ class TestDelegateClass < Test::Unit::TestCase
     assert_equal(:m, obj.m, "[ruby-dev:33116]")
   end
 
+  def test_delegate_class_block
+    klass = DelegateClass(Array) do
+      alias foo first
+    end
+    assert_equal(1, klass.new([1]).foo)
+  end
+
   def test_systemcallerror_eq
     e = SystemCallError.new(0)
     assert((SimpleDelegator.new(e) == e) == (e == SimpleDelegator.new(e)), "[ruby-dev:34808]")
@@ -78,6 +85,37 @@ class TestDelegateClass < Test::Unit::TestCase
     assert_equal(:m, bar.send(:delegate_test_m), bug)
   end
 
+  class Parent
+    def parent_public; end
+
+    protected
+
+    def parent_protected; end
+  end
+
+  class Child < DelegateClass(Parent)
+  end
+
+  class Parent
+    def parent_public_added; end
+
+    protected
+
+    def parent_protected_added; end
+  end
+
+  def test_public_instance_methods
+    ignores = Object.public_instance_methods | Delegator.public_instance_methods
+    assert_equal([:parent_public, :parent_public_added], (Child.public_instance_methods - ignores).sort)
+    assert_equal([:parent_public, :parent_public_added], (Child.new(Parent.new).public_methods - ignores).sort)
+  end
+
+  def test_protected_instance_methods
+    ignores = Object.protected_instance_methods | Delegator.protected_instance_methods
+    assert_equal([:parent_protected, :parent_protected_added], (Child.protected_instance_methods - ignores).sort)
+    assert_equal([:parent_protected, :parent_protected_added], (Child.new(Parent.new).protected_methods - ignores).sort)
+  end
+
   class IV < DelegateClass(Integer)
     attr_accessor :var
 
@@ -139,6 +177,25 @@ class TestDelegateClass < Test::Unit::TestCase
     assert_not_operator(s0, :eql?, "bar")
   end
 
+  def test_keyword_and_hash
+    foo = Object.new
+    def foo.bar(*args)
+      args
+    end
+    def foo.foo(*args, **kw)
+      [args, kw]
+    end
+    d = SimpleDelegator.new(foo)
+    assert_equal([[], {}], d.foo)
+    assert_equal([], d.bar)
+    assert_equal([[], {:a=>1}], d.foo(:a=>1))
+    assert_equal([{:a=>1}], d.bar(:a=>1))
+    assert_warn(/Using the last argument as keyword parameters is deprecated.*The called method `foo'/m) do
+      assert_equal([[], {:a=>1}], d.foo({:a=>1}))
+    end
+    assert_equal([{:a=>1}], d.bar({:a=>1}))
+  end
+
   class Foo
     private
     def delegate_test_private
@@ -258,4 +315,19 @@ class TestDelegateClass < Test::Unit::TestCase
   def test_dir_in_delegator_class
     assert_equal(__dir__, Bug9403::DC.dir_name, Bug9403::Name)
   end
+
+  def test_module_methods_vs_kernel_methods
+    delegate = SimpleDelegator.new(Object.new)
+    assert_raise(NoMethodError) do
+      delegate.constants
+    end
+  end
+
+  def test_basicobject
+    o = BasicObject.new
+    def o.bar; 1; end
+    delegate = SimpleDelegator.new(o)
+    assert_equal(1, delegate.bar)
+    assert_raise(NoMethodError, /undefined method `foo' for/) { delegate.foo }
+  end
 end
