@@ -4,14 +4,6 @@
 static VALUE rb_cMutex, rb_cQueue, rb_cSizedQueue, rb_cConditionVariable;
 static VALUE rb_eClosedQueueError;
 
-/*
- * keep these globally so we can walk and reinitialize them at fork
- * in the child process
- */
-static LIST_HEAD(szqueue_list);
-static LIST_HEAD(queue_list);
-static LIST_HEAD(condvar_list);
-
 /* sync_waiter is always on-stack */
 struct sync_waiter {
     rb_thread_t *th;
@@ -20,40 +12,37 @@ struct sync_waiter {
 
 #define MUTEX_ALLOW_TRAP FL_USER1
 
-static int
-wakeup_one(struct list_head *head)
+static void
+sync_wakeup(struct list_head *head, long max)
 {
-    struct sync_waiter *cur = 0, *next = 0;
+    struct sync_waiter *cur = 0, *next;
 
     list_for_each_safe(head, cur, next, node) {
 	list_del_init(&cur->node);
 	if (cur->th->status != THREAD_KILLED) {
 	    rb_threadptr_interrupt(cur->th);
 	    cur->th->status = THREAD_RUNNABLE;
-	    return TRUE;
+	    if (--max == 0) return;
 	}
     }
-    return FALSE;
+}
+
+static void
+wakeup_one(struct list_head *head)
+{
+    sync_wakeup(head, 1);
 }
 
 static void
 wakeup_all(struct list_head *head)
 {
-    struct sync_waiter *cur = 0, *next = 0;
-
-    list_for_each_safe(head, cur, next, node) {
-	list_del_init(&cur->node);
-	if (cur->th->status != THREAD_KILLED) {
-	    rb_threadptr_interrupt(cur->th);
-	    cur->th->status = THREAD_RUNNABLE;
-	}
-    }
+    sync_wakeup(head, LONG_MAX);
 }
 
 /* Mutex */
 
 typedef struct rb_mutex_struct {
-    struct rb_thread_struct volatile *th;
+    rb_thread_t *th;
     struct rb_mutex_struct *next_mutex;
     struct list_head waitq; /* protected by GVL */
 } rb_mutex_t;
@@ -63,7 +52,7 @@ static void rb_mutex_abandon_all(rb_mutex_t *mutexes);
 static void rb_mutex_abandon_keeping_mutexes(rb_thread_t *th);
 static void rb_mutex_abandon_locking_mutex(rb_thread_t *th);
 #endif
-static const char* rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t volatile *th);
+static const char* rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t *th);
 
 /*
  *  Document-class: Mutex
@@ -89,9 +78,6 @@ static const char* rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t volatile *t
  *
  */
 
-#define GetMutexPtr(obj, tobj) \
-    TypedData_Get_Struct((obj), rb_mutex_t, &mutex_data_type, (tobj))
-
 #define mutex_mark NULL
 
 static size_t
@@ -131,6 +117,16 @@ static const rb_data_type_t mutex_data_type = {
     0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
+static rb_mutex_t *
+mutex_ptr(VALUE obj)
+{
+    rb_mutex_t *mutex;
+
+    TypedData_Get_Struct(obj, rb_mutex_t, &mutex_data_type, mutex);
+
+    return mutex;
+}
+
 VALUE
 rb_obj_is_mutex(VALUE obj)
 {
@@ -180,16 +176,15 @@ rb_mutex_new(void)
 VALUE
 rb_mutex_locked_p(VALUE self)
 {
-    rb_mutex_t *mutex;
-    GetMutexPtr(self, mutex);
+    rb_mutex_t *mutex = mutex_ptr(self);
+
     return mutex->th ? Qtrue : Qfalse;
 }
 
 static void
 mutex_locked(rb_thread_t *th, VALUE self)
 {
-    rb_mutex_t *mutex;
-    GetMutexPtr(self, mutex);
+    rb_mutex_t *mutex = mutex_ptr(self);
 
     if (th->keeping_mutexes) {
 	mutex->next_mutex = th->keeping_mutexes;
@@ -207,9 +202,8 @@ mutex_locked(rb_thread_t *th, VALUE self)
 VALUE
 rb_mutex_trylock(VALUE self)
 {
-    rb_mutex_t *mutex;
+    rb_mutex_t *mutex = mutex_ptr(self);
     VALUE locked = Qfalse;
-    GetMutexPtr(self, mutex);
 
     if (mutex->th == 0) {
 	rb_thread_t *th = GET_THREAD();
@@ -229,12 +223,22 @@ rb_mutex_trylock(VALUE self)
  */
 static const rb_thread_t *patrol_thread = NULL;
 
+static VALUE
+mutex_owned_p(rb_thread_t *th, rb_mutex_t *mutex)
+{
+    if (mutex->th == th) {
+        return Qtrue;
+    }
+    else {
+        return Qfalse;
+    }
+}
+
 static VALUE
 do_mutex_lock(VALUE self, int interruptible_p)
 {
     rb_thread_t *th = GET_THREAD();
-    rb_mutex_t *mutex;
-    GetMutexPtr(self, mutex);
+    rb_mutex_t *mutex = mutex_ptr(self);
 
     /* When running trap handler */
     if (!FL_TEST_RAW(self, MUTEX_ALLOW_TRAP) &&
@@ -253,8 +257,8 @@ do_mutex_lock(VALUE self, int interruptible_p)
 
 	while (mutex->th != th) {
 	    enum rb_thread_status prev_status = th->status;
-	    struct timeval *timeout = 0;
-	    struct timeval tv = { 0, 100000 }; /* 100ms */
+	    rb_hrtime_t *timeout = 0;
+	    rb_hrtime_t rel = rb_msec2hrtime(100);
 
 	    th->status = THREAD_STOPPED_FOREVER;
 	    th->locking_mutex = self;
@@ -266,7 +270,7 @@ do_mutex_lock(VALUE self, int interruptible_p)
 	     */
 	    if ((vm_living_thread_num(th->vm) == th->vm->sleeper) &&
 		!patrol_thread) {
-		timeout = &tv;
+		timeout = &rel;
 		patrol_thread = th;
 	    }
 
@@ -299,11 +303,16 @@ do_mutex_lock(VALUE self, int interruptible_p)
                     mutex->th = th;
                     mutex_locked(th, self);
                 }
-            } else {
+            }
+            else {
                 if (mutex->th == th) mutex_locked(th, self);
             }
 	}
     }
+
+    // assertion
+    if (mutex_owned_p(th, mutex) == Qfalse) rb_bug("do_mutex_lock: mutex is not owned.");
+
     return self;
 }
 
@@ -335,20 +344,14 @@ rb_mutex_lock(VALUE self)
 VALUE
 rb_mutex_owned_p(VALUE self)
 {
-    VALUE owned = Qfalse;
     rb_thread_t *th = GET_THREAD();
-    rb_mutex_t *mutex;
+    rb_mutex_t *mutex = mutex_ptr(self);
 
-    GetMutexPtr(self, mutex);
-
-    if (mutex->th == th)
-	owned = Qtrue;
-
-    return owned;
+    return mutex_owned_p(th, mutex);
 }
 
 static const char *
-rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t volatile *th)
+rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t *th)
 {
     const char *err = NULL;
 
@@ -359,8 +362,8 @@ rb_mutex_unlock_th(rb_mutex_t *mutex, rb_thread_t volatile *th)
 	err = "Attempt to unlock a mutex which is locked by another thread";
     }
     else {
-	struct sync_waiter *cur = 0, *next = 0;
-	rb_mutex_t *volatile *th_mutex = &th->keeping_mutexes;
+	struct sync_waiter *cur = 0, *next;
+	rb_mutex_t **th_mutex = &th->keeping_mutexes;
 
 	mutex->th = 0;
 	list_for_each_safe(&mutex->waitq, cur, next, node) {
@@ -400,8 +403,7 @@ VALUE
 rb_mutex_unlock(VALUE self)
 {
     const char *err;
-    rb_mutex_t *mutex;
-    GetMutexPtr(self, mutex);
+    rb_mutex_t *mutex = mutex_ptr(self);
 
     err = rb_mutex_unlock_th(mutex, GET_THREAD());
     if (err) rb_raise(rb_eThreadError, "%s", err);
@@ -413,23 +415,19 @@ rb_mutex_unlock(VALUE self)
 static void
 rb_mutex_abandon_keeping_mutexes(rb_thread_t *th)
 {
-    if (th->keeping_mutexes) {
-	rb_mutex_abandon_all(th->keeping_mutexes);
-    }
+    rb_mutex_abandon_all(th->keeping_mutexes);
     th->keeping_mutexes = NULL;
 }
 
 static void
 rb_mutex_abandon_locking_mutex(rb_thread_t *th)
 {
-    rb_mutex_t *mutex;
+    if (th->locking_mutex) {
+        rb_mutex_t *mutex = mutex_ptr(th->locking_mutex);
 
-    if (!th->locking_mutex) return;
-
-    GetMutexPtr(th->locking_mutex, mutex);
-    if (mutex->th == th)
-	rb_mutex_abandon_all(mutex);
-    th->locking_mutex = Qfalse;
+        list_head_init(&mutex->waitq);
+        th->locking_mutex = Qfalse;
+    }
 }
 
 static void
@@ -445,20 +443,6 @@ rb_mutex_abandon_all(rb_mutex_t *mutexes)
 	list_head_init(&mutex->waitq);
     }
 }
-
-/*
- * All other threads are dead in the a new child process, so waitqs
- * contain references to dead threads which we need to clean up
- */
-static void
-rb_mutex_cleanup_keeping_mutexes(const rb_thread_t *current_thread)
-{
-    rb_mutex_t *mutex = current_thread->keeping_mutexes;
-    while (mutex) {
-        list_head_init(&mutex->waitq);
-        mutex = mutex->next_mutex;
-    }
-}
 #endif
 
 static VALUE
@@ -471,8 +455,9 @@ rb_mutex_sleep_forever(VALUE time)
 static VALUE
 rb_mutex_wait_for(VALUE time)
 {
-    struct timeval *t = (struct timeval *)time;
-    sleep_timeval(GET_THREAD(), *t, 0); /* permit spurious check */
+    rb_hrtime_t *rel = (rb_hrtime_t *)time;
+    /* permit spurious check */
+    sleep_hrtime(GET_THREAD(), *rel, 0);
     return Qnil;
 }
 
@@ -485,13 +470,17 @@ rb_mutex_sleep(VALUE self, VALUE timeout)
     if (!NIL_P(timeout)) {
         t = rb_time_interval(timeout);
     }
+
     rb_mutex_unlock(self);
     beg = time(0);
     if (NIL_P(timeout)) {
 	rb_ensure(rb_mutex_sleep_forever, Qnil, mutex_lock_uninterruptible, self);
     }
     else {
-	rb_ensure(rb_mutex_wait_for, (VALUE)&t, mutex_lock_uninterruptible, self);
+        rb_hrtime_t rel = rb_timeval2hrtime(&t);
+
+        rb_ensure(rb_mutex_wait_for, (VALUE)&rel,
+                  mutex_lock_uninterruptible, self);
     }
     RUBY_VM_CHECK_INTS_BLOCKING(GET_EC());
     end = time(0) - beg;
@@ -517,7 +506,7 @@ mutex_sleep(int argc, VALUE *argv, VALUE self)
 {
     VALUE timeout;
 
-    rb_scan_args(argc, argv, "01", &timeout);
+    timeout = rb_check_arity(argc, 0, 1) ? argv[0] : Qnil;
     return rb_mutex_sleep(self, timeout);
 }
 
@@ -544,7 +533,7 @@ rb_mutex_synchronize(VALUE mutex, VALUE (*func)(VALUE arg), VALUE arg)
  * completes.  See the example under +Mutex+.
  */
 static VALUE
-rb_mutex_synchronize_m(VALUE self, VALUE args)
+rb_mutex_synchronize_m(VALUE self)
 {
     if (!rb_block_given_p()) {
 	rb_raise(rb_eThreadError, "must be called with a block");
@@ -714,6 +703,15 @@ queue_closed_p(VALUE self)
     return FL_TEST_RAW(self, QUEUE_CLOSED) != 0;
 }
 
+/*
+ *  Document-class: ClosedQueueError
+ *
+ *  The exception class which will be raised when pushing into a closed
+ *  Queue.  See Queue#close and SizedQueue#close.
+ */
+
+NORETURN(static void raise_closed_queue_error(VALUE self));
+
 static void
 raise_closed_queue_error(VALUE self)
 {
@@ -758,6 +756,8 @@ queue_closed_result(VALUE self, struct rb_queue *q)
  *	  end
  *	end
  *
+ *	consumer.join
+ *
  */
 
 /*
@@ -799,10 +799,12 @@ queue_do_push(VALUE self, struct rb_queue *q, VALUE obj)
  *
  * - +close+ will be ignored.
  *
- * - calling enq/push/<< will raise an exception.
+ * - calling enq/push/<< will raise a +ClosedQueueError+.
  *
  * - when +empty?+ is false, calling deq/pop/shift will return an object
  *   from the queue as usual.
+ * - when +empty?+ is true, deq(false) will not suspend the thread and will return nil.
+ *   deq(true) will raise a +ThreadError+.
  *
  * ClosedQueueError is inherited from StopIteration, so that you can break loop block.
  *
@@ -917,7 +919,7 @@ queue_do_pop(VALUE self, struct rb_queue *q, int should_block)
 
 	    qw.w.th = GET_THREAD();
 	    qw.as.q = q;
-	    list_add_tail(&qw.as.q->waitq, &qw.w.node);
+	    list_add_tail(queue_waitq(qw.as.q), &qw.w.node);
 	    qw.as.q->num_waiting++;
 
 	    rb_ensure(queue_sleep, self, queue_sleep_done, (VALUE)&qw);
@@ -1109,9 +1111,7 @@ rb_szqueue_max_set(VALUE self, VALUE vmax)
 	diff = max - sq->max;
     }
     sq->max = max;
-    while (diff-- > 0 && wakeup_one(szqueue_pushq(sq))) {
-	/* keep waking more up */
-    }
+    sync_wakeup(szqueue_pushq(sq), diff);
     return vmax;
 }
 
@@ -1224,6 +1224,15 @@ rb_szqueue_clear(VALUE self)
     return self;
 }
 
+/*
+ * Document-method: SizedQueue#length
+ * call-seq:
+ *   length
+ *   size
+ *
+ * Returns the length of the queue.
+ */
+
 static VALUE
 rb_szqueue_length(VALUE self)
 {
@@ -1346,7 +1355,7 @@ condvar_alloc(VALUE klass)
 static VALUE
 rb_condvar_initialize(VALUE self)
 {
-    struct rb_condvar *cv = condvar_ptr(self);;
+    struct rb_condvar *cv = condvar_ptr(self);
     list_head_init(&cv->waitq);
     return self;
 }
@@ -1366,8 +1375,9 @@ do_sleep(VALUE args)
 }
 
 static VALUE
-delete_from_waitq(struct sync_waiter *w)
+delete_from_waitq(VALUE v)
 {
+    struct sync_waiter *w = (void *)v;
     list_del(&w->node);
 
     return Qnil;
@@ -1387,14 +1397,11 @@ static VALUE
 rb_condvar_wait(int argc, VALUE *argv, VALUE self)
 {
     struct rb_condvar *cv = condvar_ptr(self);
-    VALUE mutex, timeout;
     struct sleep_call args;
     struct sync_waiter w;
 
-    rb_scan_args(argc, argv, "11", &mutex, &timeout);
+    rb_scan_args(argc, argv, "11", &args.mutex, &args.timeout);
 
-    args.mutex   = mutex;
-    args.timeout = timeout;
     w.th = GET_THREAD();
     list_add_tail(&cv->waitq, &w.node);
     rb_ensure(do_sleep, (VALUE)&args, delete_from_waitq, (VALUE)&w);
@@ -1435,7 +1442,7 @@ static VALUE
 undumpable(VALUE obj)
 {
     rb_raise(rb_eTypeError, "can't dump %"PRIsVALUE, rb_obj_class(obj));
-    UNREACHABLE;
+    UNREACHABLE_RETURN(Qnil);
 }
 
 static VALUE
@@ -1449,6 +1456,7 @@ define_thread_class(VALUE outer, const char *name, VALUE super)
 static void
 Init_thread_sync(void)
 {
+#undef rb_intern
 #if 0
     rb_cMutex = rb_define_class("Mutex", rb_cObject); /* teach rdoc Mutex */
     rb_cConditionVariable = rb_define_class("ConditionVariable", rb_cObject); /* teach rdoc ConditionVariable */
