@@ -2,7 +2,7 @@
 
   thread_win32.h -
 
-  $Author: nobu $
+  $Author$
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -12,8 +12,6 @@
 #ifndef RUBY_THREAD_WIN32_H
 #define RUBY_THREAD_WIN32_H
 
-#include <windows.h>
-
 # ifdef __CYGWIN__
 # undef _WIN32
 # endif
@@ -21,17 +19,10 @@
 WINBASEAPI BOOL WINAPI
 TryEnterCriticalSection(IN OUT LPCRITICAL_SECTION lpCriticalSection);
 
-typedef HANDLE rb_thread_id_t;
-
-typedef union rb_thread_lock_union {
-    HANDLE mutex;
-    CRITICAL_SECTION crit;
-} rb_thread_lock_t;
-
 typedef struct rb_thread_cond_struct {
     struct cond_event_entry *next;
     struct cond_event_entry *prev;
-} rb_thread_cond_t;
+} rb_nativethread_cond_t;
 
 typedef struct native_thread_data_struct {
     HANDLE interrupt_event;
